from os import PathLike

import numpy
import pytest
from ewoksorange.tests.utils import execute_task

from orangecontrib.ewoksxrpd.diagnose_integrate1d import OWDiagnoseIntegrate1D
from orangecontrib.ewoksxrpd.integrate_singlepattern import OWIntegrateSinglePattern

from ..tasks.integrate import IntegrateSinglePattern
from .xrpd_theory import AzimuthalPattern
from .xrpd_theory import IntensityPattern
from .xrpd_theory import Measurement
from .xrpd_theory import RadialPattern
from .xrpd_theory import Setup


@pytest.mark.parametrize("monitor_as_list", [False, True], ids=["scalar", "list"])
def test_integrate1d_task(
    tmpdir: PathLike,
    imageSetup1SampleA: Measurement,
    setup1: Setup,
    xSampleA: RadialPattern,
    ySampleA: IntensityPattern,
    monitor_as_list: bool,
):
    assert_integrate1d(
        imageSetup1SampleA,
        setup1,
        xSampleA,
        ySampleA,
        tmpdir,
        None,
        {"method": "cython"},
        monitor_as_list=monitor_as_list,
    )


def test_integrate1d_widget(
    tmpdir: PathLike,
    imageSetup1SampleA: Measurement,
    setup1: Setup,
    xSampleA: RadialPattern,
    ySampleA: IntensityPattern,
    qtapp,
):
    assert_integrate1d(
        imageSetup1SampleA,
        setup1,
        xSampleA,
        ySampleA,
        tmpdir,
        qtapp,
        {"method": "cython"},
    )


# from pyFAI.method_registry import IntegrationMethod
# for method in IntegrationMethod._registry:
#    print(f"{method.split}_{method.algo}_{method.impl}")
#
# {split}_{algo}_{impl}{target}
# split: "no", "bbox", "pseudo", "full"
# algo: "histogram", "lut", "csr"
# impl: "python", "cython", "opencl"


def test_sigma_clip_task(
    tmpdir: PathLike,
    imageSetup1SampleA: Measurement,
    setup1: Setup,
    xSampleA: RadialPattern,
    ySampleA: IntensityPattern,
):
    integration_options = {
        "error_model": "azimuthal",
        "method": "no_csr_cython",
        "integrator_name": "sigma_clip_ng",
        "extra_options": {"max_iter": 3, "thres": 0},
    }
    assert_integrate1d(
        imageSetup1SampleA,
        setup1,
        xSampleA,
        ySampleA,
        tmpdir,
        None,
        integration_options,
    )


def test_sigma_clip_widget(
    tmpdir: PathLike,
    imageSetup1SampleA: Measurement,
    setup1: Setup,
    xSampleA: RadialPattern,
    ySampleA: IntensityPattern,
    qtapp,
):
    integration_options = {
        "error_model": "azimuthal",
        "method": "no_csr_cython",
        "integrator_name": "sigma_clip_ng",
        "extra_options": {"max_iter": 3, "thres": 0},
    }
    assert_integrate1d(
        imageSetup1SampleA,
        setup1,
        xSampleA,
        ySampleA,
        tmpdir,
        qtapp,
        integration_options,
    )


def test_integrate1d_reconfig(
    tmpdir: PathLike,
    imageSetup1SampleA: Measurement,
    setup1: Setup,
    imageSetup2SampleA: Measurement,
    setup2: Setup,
    xSampleA: RadialPattern,
    ySampleA: IntensityPattern,
):
    assert_integrate1d(
        imageSetup1SampleA,
        setup1,
        xSampleA,
        ySampleA,
        tmpdir,
        None,
        {"method": "cython"},
    )
    assert_integrate1d(
        imageSetup2SampleA,
        setup2,
        xSampleA,
        ySampleA,
        tmpdir,
        None,
        {"method": "cython"},
    )
    assert_integrate1d(
        imageSetup1SampleA,
        setup1,
        xSampleA,
        ySampleA,
        tmpdir,
        None,
        {"method": "cython"},
    )
    assert_integrate1d(
        imageSetup1SampleA,
        setup1,
        xSampleA,
        ySampleA,
        tmpdir,
        None,
        {"method": "cython"},
    )


def test_integrate2d(
    imageSetup1SampleC: Measurement,
    setup1: Setup,
    radialSampleC: RadialPattern,
    azimuthalSampleC: AzimuthalPattern,
    intensitySampleC: IntensityPattern,
):
    integration_options = {
        **setup1.integration_options,
        **radialSampleC.integration_options,
        **azimuthalSampleC.integration_options,
    }
    inputs = {
        "image": imageSetup1SampleC.image,
        "detector": setup1.detector,
        "geometry": setup1.geometry,
        "energy": setup1.energy,
        "integration_options": integration_options,
    }
    inputs["monitor"] = imageSetup1SampleC.monitor
    inputs["reference"] = intensitySampleC.monitor

    output_values = execute_task(IntegrateSinglePattern, inputs=inputs)

    assert output_values["radial_units"] == radialSampleC.units
    numpy.testing.assert_allclose(radialSampleC.x, output_values["radial"], rtol=1e-6)

    assert output_values["radial_units"] == radialSampleC.units
    numpy.testing.assert_allclose(
        azimuthalSampleC.x, output_values["azimuthal"], rtol=1e-1
    )
    assert output_values["azimuthal_units"] == azimuthalSampleC.units

    assert output_values["intensity"].shape == (
        azimuthalSampleC.x.size,
        radialSampleC.x.size,
    )


def assert_integrate1d(
    measurement: Measurement,
    setup: Setup,
    xpattern: RadialPattern,
    ypattern: IntensityPattern,
    tmpdir: PathLike,
    qtapp,
    integration_options: dict,
    monitor_as_list: bool = False,
):
    integration_options = {
        **setup.integration_options,
        **xpattern.integration_options,
        **integration_options,
    }
    inputs = {
        "image": measurement.image,
        "detector": setup.detector,
        "geometry": setup.geometry,
        "energy": setup.energy,
        "integration_options": integration_options,
    }
    if monitor_as_list:
        inputs["monitors"] = [measurement.monitor, 1, None, 1]
        inputs["references"] = [ypattern.monitor, 1, 1, None]
    else:
        inputs["monitor"] = measurement.monitor
        inputs["reference"] = ypattern.monitor

    output_values = execute_task(
        (
            OWIntegrateSinglePattern.ewokstaskclass
            if qtapp is None
            else OWIntegrateSinglePattern
        ),
        inputs=inputs,
    )

    assert output_values["radial_units"] == xpattern.units
    numpy.testing.assert_allclose(xpattern.x, output_values["radial"], rtol=1e-6)
    atol = ypattern.y.max() * 0.01
    numpy.testing.assert_allclose(ypattern.y, output_values["intensity"], atol=atol)

    # Set show=True to visualize the calibration results
    filename = tmpdir / "diagnose.png"
    inputs = {
        "x": output_values["radial"],
        "y": output_values["intensity"],
        "xunits": output_values["radial_units"],
        "show": False,
        "filename": str(filename),
        # "energy": setup.energy,
        # "calibrant": "LaB6"
    }
    execute_task(
        (
            OWDiagnoseIntegrate1D.ewokstaskclass
            if qtapp is None
            else OWDiagnoseIntegrate1D
        ),
        inputs=inputs,
    )
    assert filename.exists()



import os
from numbers import Number
from typing import Mapping
from typing import Optional
from typing import Tuple
from typing import Union

import fabio.edfimage
import fabio.tifimage
import h5py
import numpy
from blissdata.h5api import dynamic_hdf5
from silx.io.url import DataUrl
from typing_extensions import TypeGuard


def hdf5_url(file_name: str, data_path: str) -> str:
    if not os.path.isabs(file_name):
        file_name = os.path.abspath(file_name)
    return f"silx://{file_name}?path={data_path}"


def split_hdf5_url_parent_data_path(url: DataUrl) -> Tuple[DataUrl, str]:
    data_path = url.data_path() or "/"
    parts = [s for s in data_path.split("/")]
    if not parts:
        raise ValueError(f"{url.path()!r} needs to refer to a non-root HDF5 group")

    name = parts[-1]
    parent_data_path = "/".join(parts[:-1])
    if parent_data_path:
        parent_url = DataUrl(f"{url.file_path()}::{parent_data_path}")
    else:
        parent_url = DataUrl(url.file_path())
    return parent_url, name


def is_data(data) -> TypeGuard[Union[numpy.ndarray, Number, str, list]]:
    if isinstance(data, (numpy.ndarray, Number)):
        return True
    if isinstance(data, (str, list)) and data:
        return True
    return False


def is_same_file(filename1: str, filename2: str) -> bool:
    return os.path.abspath(os.path.normpath(filename1)) == os.path.abspath(
        os.path.normpath(filename2)
    )


def data_from_storage(data, remove_numpy=True):
    if isinstance(data, numpy.ndarray):
        if not remove_numpy:
            return data
        elif data.ndim == 0:
            return data.item()
        else:
            return data.tolist()
    elif isinstance(data, Mapping):
        return {
            k: data_from_storage(v, remove_numpy=remove_numpy)
            for k, v in data.items()
            if not k.startswith("@")
        }
    else:
        return data


def create_hdf5_link(
    parent: h5py.Group,
    link_name: str,
    target: Union[h5py.Dataset, h5py.Group],
    relative: bool = True,
    raise_on_exists: bool = False,
    overwrite: bool = False,
) -> None:
    """
    :param parent: HDF5 group in which the link will be created
    :param link_name: relative HDF5 path of the link source with respect to :code:`parent`
    :param target: absolute HDF5 path of the link target
    :param relative: determines whether the external or internal link is absolute or relative.
                     Internal links that refer upwards are not supported and will always be absolute.
    :param raise_on_exists: raise exception when :code:`link_name` already exists
    :param overwrite: Set to True to allow overwriting existing link_name
    """
    link = hdf5_link_object(
        _get_hdf5_filename(parent.file),
        parent.name,
        link_name,
        _get_hdf5_filename(target.file),
        target.name,
        relative=relative,
    )
    if link is None:
        # Link refers to itself
        return
    if overwrite and link_name in parent:
        del parent[link_name]
    if link_name in parent and not raise_on_exists:
        # Name already exists and is not necessarily equivalent to the link we want to create
        return
    parent[link_name] = link


def _get_hdf5_filename(file_obj) -> str:
    try:
        return file_obj.filename
    except AttributeError:
        # to be fixed in blissdata
        return file_obj._retry_handler.file_obj.filename


def hdf5_link_object(
    parent_filename: str,
    parent_name: str,
    link_name: str,
    target_filename: str,
    target_name: str,
    relative: bool = True,
) -> Union[None, h5py.ExternalLink, h5py.SoftLink]:
    """
    :param parent_filename: HDF5 filename in which the link will be created
    :param parent_name: absolute HDF5 group path in :code:`parent_filename`
    :param link_name: relative HDF5 path of the link source with respect to :code:`parent_name`
    :param target_filename: HDF5 filename in which the link target is located
    :param target_name: absolute HDF5 path in :code:`target_filename` of the link target
    :param relative: determines whether the external or internal link is absolute or relative.
                     Internal links that refer upwards are not supported and will always be absolute.
    :returns: Internal or external link object to be used to create the HDF5 link.
              Returns :code:`None` when the link refers to itself.
    """
    abs_parent_filename = os.path.abspath(parent_filename)
    target_is_absolute = os.path.isabs(target_filename)
    if target_is_absolute:
        abs_target_filename = target_filename
    else:
        abs_target_filename = os.path.join(
            os.path.dirname(abs_parent_filename), target_filename
        )

    # Internal link
    if os.path.normpath(abs_parent_filename) == os.path.normpath(abs_target_filename):
        link_full_name = _normalize_hdf5_item_name(parent_name, link_name)
        target_name = _normalize_hdf5_item_name(target_name)
        if link_full_name == target_name:
            # Link refers to itself
            return
        if not relative:
            return h5py.SoftLink(target_name)
        rel_target_name = os.path.relpath(target_name, parent_name)
        if ".." in rel_target_name:
            # Internal links upwards are not supported
            return h5py.SoftLink(target_name)
        return h5py.SoftLink(rel_target_name)

    # External link
    if relative:
        target_filename = os.path.relpath(
            abs_target_filename, os.path.dirname(abs_parent_filename)
        )
    else:
        target_filename = abs_target_filename
    return h5py.ExternalLink(target_filename, target_name)


def _normalize_hdf5_item_name(*parts) -> str:
    name = "/".join([s for part in parts for s in part.split("/") if s])
    return f"/{name}"


def link_bliss_scan(
    outentry: h5py.Group, bliss_scan_url: Union[str, DataUrl], **kwargs
):
    if isinstance(bliss_scan_url, str):
        bliss_scan_url = DataUrl(bliss_scan_url)
    file_path = bliss_scan_url.file_path()
    data_path = bliss_scan_url.data_path()
    out_filename = outentry.file.filename
    ext_filename = os.path.relpath(out_filename, os.path.dirname(file_path))
    if ".." in ext_filename:
        ext_filename = file_path
    with dynamic_hdf5.File(file_path, mode="r", **kwargs) as root:
        inentry = root[data_path]
        # Wait until the file has finished writing
        _ = inentry["end_time"]
        # Link to the entire group
        for groupname in ("instrument", "sample"):
            try:
                if groupname in outentry or groupname not in inentry:
                    continue
            except Exception:  # fixed by bliss PR !5435
                continue
            create_hdf5_link(outentry, groupname, inentry[groupname])

        # Link to all sub groups
        for groupname in ("measurement",):
            if groupname not in inentry:
                continue
            igroup = inentry[groupname]
            if groupname in outentry:
                ogroup = outentry[groupname]
            else:
                ogroup = outentry.create_group(groupname)
                ogroup.attrs["NX_class"] = igroup.attrs["NX_class"]
            for name in igroup.keys():
                if name in ogroup:
                    continue
                if name not in ogroup:
                    create_hdf5_link(ogroup, name, igroup[name])


def convert_to_3d(data: Union[numpy.ndarray, Number, str, list]):
    data_arr = numpy.array(data)

    if data_arr.ndim >= 3:
        return data_arr

    if data_arr.ndim == 2:
        return data_arr.reshape(1, *data_arr.shape)

    if data_arr.ndim == 1:
        return data_arr.reshape(1, 1, *data_arr.shape)

    return data_arr.reshape(1, 1, 1)


def save_image(
    data: numpy.ndarray,
    save_path: str,
    save_name: str,
    monitor_data: Optional[Number] = None,
    metadata: Optional[dict] = None,
    ext: str = "edf",
) -> str:
    normalized_data = data / monitor_data if monitor_data else data

    header = dict(
        Summed_monitor_counts=str(monitor_data) if monitor_data else "nan",
        **(metadata if metadata else {}),
    )

    ext = ext.lower()
    img_filepath = f"{save_path}/{save_name}.{ext}"
    if ext.lower() == "edf":
        Image = fabio.edfimage.EdfImage
    elif ext == "tiff" or ext == "tif":
        Image = fabio.tifimage.TifImage
    else:
        raise ValueError(f"Unsupported ext {ext}. Only supports EDF and TIFF.")

    img = Image(data=normalized_data, header=header)
    img.write(img_filepath)

    return img_filepath



import json
import os
from importlib.metadata import distribution
from pathlib import Path
from shutil import rmtree
from typing import Optional
from urllib.parse import urlparse

import pytest

from ewoksxrpd import pyfai_api


def pytest_addoption(parser):
    parser.addoption(
        "--skip-old-pyfai",
        action="store_true",
        default=False,
        help="Run only tests for latest pyFAI integration schema version",
    )
    parser.addoption(
        "--examples", action="store_true", default=False, help="generate example data"
    )


@pytest.fixture()
def ewoksxrpd_examples_path(request, tmp_path, ewoksxrpd_repo_dir) -> Path:
    if not request.config.option.examples or not ewoksxrpd_repo_dir:
        return tmp_path

    tmp_path = ewoksxrpd_repo_dir / "examples"
    rmtree(tmp_path / "data", ignore_errors=True)
    rmtree(tmp_path / "results", ignore_errors=True)
    rmtree(tmp_path / "transient", ignore_errors=True)

    return tmp_path


@pytest.fixture(scope="session")
def ewoksxrpd_repo_dir() -> Optional[Path]:
    dist = distribution("ewoksxrpd")

    # PEP 610: editable installs include direct_url.json
    try:
        data = dist.read_text("direct_url.json")
    except FileNotFoundError:
        return None

    if not data:
        return None

    info = json.loads(data)
    dir_info = info.get("dir_info", {})
    editable = bool(dir_info.get("editable"))
    if not editable:
        return None

    # Extract and resolve the filesystem path
    url = info.get("url")
    if not url:
        return None

    parsed = urlparse(url)
    if parsed.scheme != "file":
        return None

    path = Path(parsed.path)

    # Check that the directory exists and is writeable
    if path.is_dir() and os.access(path, os.W_OK):
        return path

    return None


_PYFAI_INTEGRATION_VERSIONS = [
    None,
    *range(
        pyfai_api.INTEGRATION_SCHEMA_VERSION_MIN,
        pyfai_api.INTEGRATION_SCHEMA_VERSION_MAX + 1,
    ),
]


@pytest.fixture(
    scope="session",
    params=_PYFAI_INTEGRATION_VERSIONS,
    ids=[f"v{v}" for v in _PYFAI_INTEGRATION_VERSIONS],
)
def pyfai_integration_version(request: pytest.FixtureRequest):
    """Fixture that provides all possible integration schema versions."""
    if request.config.getoption("--skip-old-pyfai") and request.param is not None:
        request.applymarker(
            pytest.mark.skip(
                reason="Skipping old pyFAI versions since --skip-old-pyfai was given"
            )
        )
    return request.param



import time
from contextlib import ExitStack
from typing import Any
from typing import Generator
from typing import Iterator
from typing import Optional
from typing import Tuple

import h5py
from blissdata.h5api import dynamic_hdf5
from ewokscore import Task

from .data_access import TaskWithDataAccess
from .utils.data_utils import hdf5_url
from .utils.data_utils import is_same_file
from .utils.sum_utils import ImageSum
from .utils.sum_utils import generate_range
from .utils.sum_utils import save_sum

SUM_TYPES = ["per_scan", "all_scans", "both"]


class SumBlissScanImages(
    TaskWithDataAccess,
    input_names=["filename", "scan", "detector_name", "output_filename"],
    optional_input_names=[
        "monitor_name",
        "subscan",
        "scan_memory_url",
        "output_process",
        "background_step",
    ],
    output_names=["output_uri", "monitor"],
):
    """Sum images of a single camera of a single Bliss scan"""

    def run(self):
        with ExitStack() as stack:
            self._run_in_stack(stack)

    def _run_in_stack(self, stack: ExitStack):
        filename: str = self.inputs.filename
        scan: int = self.inputs.scan
        subscan = self.get_input_value("subscan", 1)
        detector_name: str = self.inputs.detector_name
        output_filename: str = self.inputs.output_filename
        output_nxentry_url = hdf5_url(output_filename, f"/{scan}.{subscan}")
        monitor_name: Optional[str] = self.get_input_value("monitor_name", None)
        output_nxprocess_name: str = self.get_input_value("output_process", "sum")
        background_step: bool = self.get_input_value("background_step", 1)

        if self.inputs.scan_memory_url:
            raise NotImplementedError("data from memory not supported yet")
        else:
            lima_iterator_context = self.iter_bliss_data(
                filename,
                scan,
                lima_names=[detector_name],
                subscan=subscan,
            )
            lima_iterator = stack.enter_context(lima_iterator_context)

            with dynamic_hdf5.File(
                filename, lima_names=[detector_name], **self.get_retry_options()
            ) as root:
                data_name = f"{scan}.{subscan}/instrument/{detector_name}/data"
                # getting the shape should block until the end of the scan
                dataset_shape = root[data_name].shape
                nb_points_in_scan = dataset_shape[0]

            if monitor_name:
                counter_iterator_context = self.iter_bliss_data(
                    filename,
                    scan,
                    counter_names=[monitor_name],
                    lima_names=[],
                    subscan=subscan,
                )
                counter_iterator = stack.enter_context(counter_iterator_context)
            else:
                counter_iterator = (
                    {monitor_name: None} for _ in range(nb_points_in_scan)
                )

        with self.open_h5item(output_nxentry_url, mode="a", create=True) as nxentry:
            nxprocess = nxentry.create_group(output_nxprocess_name)
            nxprocess.attrs["NX_class"] = "NXprocess"
            nxprocess.attrs["default"] = "results"
            nxdata = nxprocess.create_group("results")
            nxdata.attrs["NX_class"] = "NXdata"

            scan_sum = None
            summed_indices = []

            tstart = time.time()
            for scan_index, ctrdata in iterate_scan_with_skip(
                counter_iterator, background_step
            ):
                limadata = next(lima_iterator)

                image = limadata[detector_name]
                if scan_sum is None:
                    scan_sum = ImageSum(image.shape)

                scan_sum.add_to_sum(image, monitor=ctrdata[monitor_name])
                summed_indices.append(scan_index)

            if scan_sum is None:
                tend = time.time()
                raise RuntimeError(
                    f"No scan data yielded within {tend-tstart:03f} seconds from {filename}::/{scan}.{subscan}"
                )

            image_dset, monitor_dset = save_sum(
                nxdata,
                name=f"Scan{scan}-Images{summed_indices[0]}-{summed_indices[-1]}",
                image_sum=scan_sum,
            )

            self.outputs.output_uri = f"{nxdata.file.filename}::{image_dset.name}"
            if monitor_dset is not None:
                self.outputs.monitor = scan_sum.summed_monitor
            else:
                self.outputs.monitor = None


def iterate_scan_with_skip(
    scan_iterator: Iterator[Any], step: int
) -> Generator[Tuple[int, Any], None, None]:
    """
    :param scan_iterator: iterate over all scan points
    :param step: -1 means no skipping, 0 means skipping the first, N>0 means skipping every N points starting from 0
    :yields: the non-skipped scan index and corresponding data
    """
    if step < 0:  # no skipping
        for scan_index, data in enumerate(scan_iterator):
            yield scan_index, data
    elif step == 0:  # skip the first
        for scan_index, data in enumerate(scan_iterator):
            if scan_index == 0:
                continue
            yield scan_index, data
    else:  # skip every N points, starting from zero
        block_size = step + 1
        for scan_index, data in enumerate(scan_iterator):
            if (scan_index % block_size) == 0:
                continue
            yield scan_index, data


class SumImages(
    Task,
    input_names=["filename", "detector_name", "output_filename"],
    optional_input_names=[
        "start_scan",
        "end_scan",
        "start_image",
        "end_image",
        "block_size",
        "monitor_name",
        "output_entry",
        "output_process",
        "sum_type",
    ],
    output_names=["output_uris", "monitor_uris"],
):
    """Sum images of a single camera from a Bliss scan file

    For each scan, images are added in blocks of `block_size` images (one block with all images by default).

    The result contains:
        * the block sums when sum_type=per_scan or sum_type=both
        * the sum of the block sums when sum_type=all_scans or sum_type=both
    """

    def run(self):
        filename: str = self.inputs.filename
        detector_name: str = self.inputs.detector_name
        output_filename: str = self.inputs.output_filename
        start_scan: int = self.get_input_value("start_scan", 1)
        end_scan: Optional[int] = self.get_input_value("end_scan", None)
        start_image: int = self.get_input_value("start_image", 0)
        end_image: Optional[int] = self.get_input_value("end_image", None)
        block_size: Optional[int] = self.get_input_value("block_size", None)
        monitor_name: Optional[str] = self.get_input_value("monitor_name", None)
        output_entry: str = self.get_input_value("output_entry", "processing")
        output_process: str = self.get_input_value("output_process", "sum")
        sum_type: str = self.get_input_value("sum_type", "per_scan")

        if sum_type not in SUM_TYPES:
            raise TypeError(
                f"sum_type must be one of the following values: {SUM_TYPES}. Got {sum_type} instead."
            )
        save_scan_sums = sum_type == "per_scan" or sum_type == "both"
        save_full_sum = sum_type == "all_scans" or sum_type == "both"

        if is_same_file(filename, output_filename):
            mode = "a"
        else:
            mode = "r"

        with h5py.File(filename, mode=mode) as h5infile:
            nscans = len(h5infile)
            scan_range = generate_range(start_scan, end_scan, nscans + 1)

            first_scan_name = list(h5infile.keys())[0]
            nimages, *detector_shape = h5infile[
                f"{first_scan_name}/measurement/{detector_name}"
            ].shape
            image_range = list(generate_range(start_image, end_image, nimages))

            scan_sum = ImageSum(detector_shape)
            full_sum = ImageSum(detector_shape)

            with h5py.File(output_filename, mode="a") as h5outfile:
                out_entry = h5outfile.require_group(output_entry)
                out_entry.attrs.setdefault("NX_class", "NXentry")
                out_entry.attrs["default"] = output_process
                out_process = out_entry.create_group(output_process)
                out_process.attrs["NX_class"] = "NXprocess"
                out_process.attrs["default"] = "results"
                out_results = out_process.create_group("results")
                out_results.attrs["NX_class"] = "NXdata"

                output_uris: list[str] = []
                monitor_uris: list[str] = []
                for scan_number in scan_range:
                    scan_images = h5infile[
                        f"{scan_number}.1/measurement/{detector_name}"
                    ]
                    scan_monitor = (
                        h5infile[f"{scan_number}.1/measurement/{monitor_name}"]
                        if monitor_name
                        else None
                    )
                    assert isinstance(scan_images, h5py.Dataset)
                    if scan_monitor is not None:
                        assert isinstance(scan_monitor, h5py.Dataset)

                    for image_number in image_range:
                        image = scan_images[image_number]
                        monitor = scan_monitor[image_number] if scan_monitor else None
                        scan_sum.add_to_sum(image, monitor)
                        full_sum.add_to_sum(image, monitor)

                        if save_scan_sums and scan_sum.nb_images == block_size:
                            name = f"Scan{scan_number}-Images{image_number - scan_sum.nb_images + 1}-{image_number}"
                            image_dset, monitor_dset = save_sum(
                                out_results, name=name, image_sum=scan_sum
                            )
                            output_uris.append(f"{output_filename}::{image_dset.name}")
                            if monitor_dset is not None:
                                monitor_uris.append(
                                    f"{output_filename}::{monitor_dset.name}"
                                )

                            # Move to next sum
                            scan_sum.reset()

                    if save_scan_sums and scan_sum.nb_images > 0:
                        name = f"Scan{scan_number}-Images{image_number - scan_sum.nb_images + 1}-{image_number}"
                        image_dset, monitor_dset = save_sum(
                            out_results, name=name, image_sum=scan_sum
                        )
                        output_uris.append(f"{output_filename}::{image_dset.name}")
                        if monitor_dset is not None:
                            monitor_uris.append(
                                f"{output_filename}::{monitor_dset.name}"
                            )

                if save_full_sum:
                    name = f"Sum of scans {scan_range.start} to {scan_range.stop - scan_range.step}"
                    image_dset, monitor_dset = save_sum(
                        out_results, name=name, image_sum=full_sum
                    )
                    output_uris.append(f"{output_filename}::{image_dset.name}")
                    if monitor_dset is not None:
                        monitor_uris.append(f"{output_filename}::{monitor_dset.name}")

        self.outputs.output_uris = output_uris
        if monitor_uris:
            self.outputs.monitor_uris = monitor_uris
        else:
            self.outputs.monitor_uris = None



import enum
import importlib.metadata
import json
import logging
import re
import warnings
from collections.abc import Mapping
from collections.abc import Sequence
from typing import Any
from typing import Dict
from typing import Iterator
from typing import List
from typing import NamedTuple
from typing import Optional
from typing import Tuple
from typing import Union

import h5py
import numpy
from ewokscore.missing_data import is_missing_data
from ewokscore.task import TaskInputError
from pyFAI.average import average_images

try:
    from pyFAI.integrator.azimuthal import AzimuthalIntegrator
except ImportError:
    from pyFAI import AzimuthalIntegrator
from pyFAI.io.ponifile import PoniFile
from pyFAI.units import Unit
from silx.io.dictdump import dicttonx
from silx.io.url import DataUrl

from .data_utils import create_hdf5_link
from .data_utils import data_from_storage
from .xrpd_utils import energy_wavelength

logger = logging.getLogger(__name__)

MULTIGEOMETRY_FIRST_PONI = "multigeometry_first_poni"
_PYFAI_VERSION = importlib.metadata.version("pyFAI")


_REPLACE_PATTERNS = {
    "/gpfs/[^/]+/": "/",
    "/mnt/multipath-shares/": "/",
    "/lbsram/": "/",
}


class AxisInfo(NamedTuple):
    name: str
    units: str

    def to_str(self):
        return f"{self.name}_{self.units}"


def parse_pyfai_units(
    pyfai_unit: Union[Unit, Tuple[Unit, Unit]],
) -> Tuple[AxisInfo, AxisInfo]:
    """
    Parse PyFAI result units into a tuple containing the radial axis info and the azimuthal axis info.
    Handles pyFAI >= 2024.1 units (tuple with radial+azimuthal) and pyFAI < 2024.1 units (only radial unit).

    If no info is available for the azimuthal axis, it will default to name "chi" and unit "deg".
    """
    if isinstance(pyfai_unit, tuple):
        return _parse_pyfai_unit(pyfai_unit[0]), _parse_pyfai_unit(pyfai_unit[1])

    return _parse_pyfai_unit(pyfai_unit), AxisInfo("chi", "deg")


def _parse_pyfai_unit(pyfai_unit: Unit) -> AxisInfo:
    unit_tuple = tuple(pyfai_unit.name.split("_"))
    if len(unit_tuple) != 2:
        raise ValueError(f"Expected unit to be of the form X_Y. Got {pyfai_unit.name}")

    return AxisInfo(*unit_tuple)


def parse_string_units(unit: Union[str, numpy.ndarray]) -> AxisInfo:
    if isinstance(unit, numpy.ndarray):
        unit = unit.item()
    if not isinstance(unit, str):
        raise TypeError(type(unit))
    unit_tuple = tuple(unit.split("_"))
    if len(unit_tuple) != 2:
        raise ValueError(f"Expected unit to be of the form X_Y. Got {unit}")
    return AxisInfo(*unit_tuple)


def read_config(
    filename: Optional[str], replace_patterns: Optional[Dict[str, str]] = None
) -> dict:
    if not filename:
        return dict()
    if filename.endswith(".json"):
        parameters = _read_json(filename)
    else:
        parameters = _read_poni(filename)
    return normalize_parameters(parameters, replace_patterns=replace_patterns)


def _read_json(filename: str) -> dict:
    with open(filename, "r") as fp:
        return json.load(fp)


def _read_poni(filename: str) -> dict:
    return PoniFile(filename).as_dict()


def normalize_parameters(
    parameters: Any, replace_patterns: Optional[Dict[str, str]] = None
) -> Any:
    if replace_patterns is None:
        replace_patterns = _REPLACE_PATTERNS
    if isinstance(parameters, str):
        for pattern, repl in replace_patterns.items():
            parameters = re.sub(pattern, repl, parameters)
        return parameters
    if isinstance(parameters, Mapping):
        return {
            k: normalize_parameters(v, replace_patterns=replace_patterns)
            for k, v in parameters.items()
        }
    if isinstance(parameters, Sequence):
        return [
            normalize_parameters(v, replace_patterns=replace_patterns)
            for v in parameters
        ]
    if isinstance(parameters, enum.Enum):
        return parameters.value
    if isinstance(parameters, (int, float)):
        return parameters
    if parameters is None:
        return None
    warnings.warn(
        f"Unexpected pyFAI configuration parameter type '{type(parameters)}'",
        UserWarning,
    )
    return parameters


def split_worker_and_integration_options(
    ewoks_pyfai_options: Mapping,
) -> Tuple[dict, dict]:
    integration_options = dict(ewoks_pyfai_options)

    # `Worker` constructor arguments
    worker_keys = (
        "integrator_name",  # "integrate1d", "sigma_clip_ng", ...
        "extra_options",  # Depends in the integrator. E.g. integrator_name="sigma_clip_ng"
        # accepts {"max_iter":3, "thres":0, "error_model": "azimuthal"}
        "shapeOut",  # Shape of the result (nbpt_rad, nbpt_azim).
        "shapeIn",  # Shape of the raw data.
        "azimuthalIntegrator",  # An AzimuthalIntegrator instance.
    )
    mixed_keys = (
        "units",  # Radial units like "2th_deg", "r_mm", "q_nm^-1" ...
        "dummy",  # To identify invalid pixels in the raw data.
        "delta_dummy",  # To identify invalid pixels in the raw data.
    )

    # Extract the arguments only meant for the `Worker`
    worker_options = {
        k: integration_options.pop(k) for k in worker_keys if k in integration_options
    }

    # Duplicate arguments which are worker and integration arguments
    for key in mixed_keys:
        value = integration_options.get(key)
        if value is not None:
            worker_options[key] = value

    # Validate worker and integration arguments
    nbpt_rad = integration_options.get("nbpt_rad") or 1024
    integration_options["nbpt_rad"] = nbpt_rad

    nbpt_azim = integration_options.get("nbpt_azim") or 1
    integration_options["nbpt_azim"] = nbpt_azim

    worker_options["shapeOut"] = (nbpt_azim, nbpt_rad)

    _ensure_finite_pair(integration_options, "radial_range_min", "radial_range_max")
    _ensure_finite_pair(integration_options, "azimuth_range_min", "azimuth_range_max")

    return worker_options, integration_options


def _ensure_finite_pair(parameters: dict, key1: str, key2: str) -> None:
    value1 = parameters.pop(key1, None)
    value2 = parameters.pop(key2, None)

    if value1 is None and value2 is None:
        pass
    elif value1 is None:
        logger.warning("Ignore %r because %r is not provided", key2, key1)
    elif value2 is None:
        logger.warning("Ignore %r because %r is not provided", key1, key2)
    elif not numpy.isfinite(value1):
        logger.warning("Ignore %r and %r because %r is not finite", key1, key2, key1)
    elif not numpy.isfinite(value2):
        logger.warning("Ignore %r and %r because %r is not finite", key2, key1, key2)
    else:
        parameters[key1] = value1
        parameters[key2] = value2


def convert_ewoks_options_to_mg_options(ewoks_pyfai_options: Mapping) -> Dict[str, Any]:
    """Convert Ewoks PyFAI options (WorkerConfig keys) to Multigeometry options (integrate2d parameters)"""
    integration_options = dict(ewoks_pyfai_options)

    npt_rad = integration_options.pop("nbpt_rad", None)
    if npt_rad:
        integration_options["npt_rad"] = npt_rad

    npt_azim = integration_options.pop("nbpt_azim", None)
    if npt_azim:
        integration_options["npt_azim"] = npt_azim

    radial_range_min = integration_options.pop("radial_range_min", None)
    radial_range_max = integration_options.pop("radial_range_max", None)
    if radial_range_min is not None and radial_range_max is not None:
        integration_options["radial_range"] = (radial_range_min, radial_range_max)

    azim_range_min = integration_options.pop("azimuth_range_min", None)
    azim_range_max = integration_options.pop("azimuth_range_max", None)
    if azim_range_min is not None and azim_range_max is not None:
        integration_options["azimuth_range"] = (azim_range_min, azim_range_max)

    return integration_options


def split_multi_geom_and_integration_options(
    ewoks_pyfai_options: Mapping,
) -> Tuple[dict, dict]:
    integration_options = convert_ewoks_options_to_mg_options(ewoks_pyfai_options)

    multi_geometry_keys = ("unit", "radial_range", "azimuth_range", "empty")

    multi_geometry_options = {
        k: integration_options.pop(k)
        for k in multi_geometry_keys
        if k in integration_options
    }

    return multi_geometry_options, integration_options


def extract_images_from_integration_options(
    integration_options: Mapping,
) -> Tuple[
    dict, Optional[numpy.ndarray], Optional[numpy.ndarray], Optional[numpy.ndarray]
]:
    """Extract all image related information from integration options and modify
    integration options accordingly.

    :return: integration options, mask image, flat-field image, dark-current image
    """
    integration_options = dict(integration_options)

    mask = _extract_mask(integration_options)
    flatfield = _extract_flatfield(integration_options)
    darkcurrent = _extract_darkcurrent(integration_options)
    darkflatmethod = integration_options.pop("darkflatmethod", None)

    if darkflatmethod is None:
        return integration_options, mask, flatfield, darkcurrent

    # Flat field correction:
    #   - default: Icor = (I - dark) / flat
    #   - counts:  Icor = (I - dark) / max(flat - dark, 1)

    flat_field = _get_image_urls(integration_options, "flat_field", "do_flat")
    if flat_field:
        flatfield = average_images(
            flat_field, filter_="mean", fformat=None, threshold=0
        )
        integration_options["flat_field"] = None
        integration_options["do_flat"] = False

    dark_current = _get_image_urls(integration_options, "dark_current", "do_dark")
    if dark_current:
        darkcurrent = average_images(
            dark_current, filter_="mean", fformat=None, threshold=0
        )
        integration_options["dark_current"] = None
        integration_options["do_dark"] = False

    if darkflatmethod == "counts":
        if flatfield is not None:
            if darkcurrent is not None:
                flatfield = flatfield - darkcurrent
            flatfield[flatfield < 1] = 1

    return integration_options, mask, flatfield, darkcurrent


def _extract_mask(integration_options: dict) -> Optional[numpy.ndarray]:
    """Integration options related to the mask:

    * mask_file: URL or None
    * do_mask: use mask_file or not
    * mask: URL, array or None (overwrites mask_file and do_mask when not None)
    """
    # Typically from a JSON configuration file
    mask_file = _get_image_urls(integration_options, "mask_file", "do_mask")
    do_mask = bool(mask_file)

    # Typically in memory from a previous calculation
    mask = integration_options.pop("mask", None)
    if isinstance(mask, str):
        mask_file = mask
        mask = None
        do_mask = True

    # Let pyFAI handle URLs (if any) and return in-memory (if any)
    integration_options["mask_file"] = mask_file
    integration_options["do_mask"] = do_mask
    return mask


def _extract_flatfield(integration_options: dict) -> Optional[numpy.ndarray]:
    """Integration options related to the flatfield:

    * flat_field: URL(s) (list of strings or comma separate string) or None
    * do_flat: use flat_field or not
    * flatfield: URL(s) (list of strings or comma separate string), array or None (overwrites flat_field and do_flat when not None)

    When a numpy array is returned, the flat field is removed from the integration options.
    """
    # Typically from a JSON configuration file
    flat_field = _get_image_urls(integration_options, "flat_field", "do_flat")
    do_flat = bool(flat_field)

    # Typically in memory from a previous calculation
    flatfield = integration_options.pop("flatfield", None)
    if isinstance(flatfield, str):
        flatfield = [s.strip() for s in flatfield.split(",")]
    if isinstance(flatfield, Sequence) and isinstance(flatfield[0], str):
        flat_field = flatfield
        flatfield = None
        do_flat = True
    if flatfield is not None:
        flat_field = None
        do_flat = False

    # Let pyFAI handle URLs (if any) and return in-memory (if any)
    integration_options["flat_field"] = flat_field
    integration_options["do_flat"] = do_flat
    return flatfield


def _extract_darkcurrent(integration_options: dict) -> Optional[numpy.ndarray]:
    """Integration options related to the dark-current:

    * dark_current: URL(s) (list of strings or comma separate string) or None
    * do_dark: use flat_field or not
    * darkcurrent: URL(s) (list of strings or comma separate string), array or None (overwrites dark_current and do_dark when not None)

    When a numpy array is returned, the dark current is removed from the integration options.
    """
    # Typically from a JSON configuration file
    dark_current = _get_image_urls(integration_options, "dark_current", "do_dark")
    do_dark = bool(dark_current)

    # Typically in memory from a previous calculation
    darkcurrent = integration_options.pop("darkcurrent", None)
    if isinstance(darkcurrent, str):
        darkcurrent = [s.strip() for s in darkcurrent.split(",")]
    if isinstance(darkcurrent, Sequence) and isinstance(darkcurrent[0], str):
        dark_current = darkcurrent
        darkcurrent = None
        do_dark = True
    if darkcurrent is not None:
        dark_current = None
        do_dark = False

    # Let pyFAI handle URLs (if any) and return in-memory (if any)
    integration_options["dark_current"] = dark_current
    integration_options["do_dark"] = do_dark
    return darkcurrent


def compile_integration_info(ewoks_pyfai_options: Mapping, **extra) -> Dict[str, Any]:
    """Compile information on a pyFAI integration process. Add and rename keys when appropriate."""
    integration_info = dict(ewoks_pyfai_options)

    mask = _extract_mask(integration_info)
    flatfield = _extract_flatfield(integration_info)
    darkcurrent = _extract_darkcurrent(integration_info)

    # Do not save the in-memory make, flat field and dark current
    if mask is not None:
        integration_info["mask_file"] = "[...]"
    if flatfield is not None:
        integration_info["flat_field"] = "[...]"
    if darkcurrent is not None:
        integration_info["dark_current"] = "[...]"

    for k, v in extra.items():
        if v is not None:
            integration_info[k] = v
    wavelength = integration_info.get("wavelength")
    if wavelength is not None:
        integration_info["energy"] = energy_wavelength(wavelength)
    return integration_info


def _get_image_urls(
    integration_options: dict, image_urls_key: str, use_image_urls_key: str
) -> Union[str, List[str], None]:
    """Return image URL(s) (list of strings or comma separate string)
    or `None` when there are no image URL(s) or when the value of
    `"use_image_urls_key"` is explicitly set to `False`.
    """
    image_urls = integration_options.get(image_urls_key, None)
    use_image = integration_options.get(use_image_urls_key, None)
    if use_image is not False and image_urls:
        # Non-empty string or list of strings
        return image_urls


def integration_info_as_text(integration_info: Mapping, **extra) -> List[str]:
    """Convert to a flat list of strings with the format `{key} = {value}`.
    Add keys and units when appropriate.
    """
    flatdict = {"pyfai_version": _PYFAI_VERSION}
    flatdict.update(integration_info)
    _add_extra(flatdict, extra)
    flatdict = dict(_flatten_dict(flatdict))

    energy = flatdict.pop("energy", None)
    if energy:
        flatdict["energy"] = f"{energy:.18e} keV"

    wavelength = flatdict.pop("wavelength", None)
    if wavelength is not None:
        flatdict["wavelength"] = f"{wavelength:.18e} m"

    geometry_dist = flatdict.pop("geometry.dist", None)
    if geometry_dist is not None:
        flatdict["distance"] = f"{geometry_dist:.18e} m"

    geometry_poni1 = flatdict.pop("geometry.poni1", None)
    if geometry_poni1 is not None:
        flatdict["center dim0"] = f"{geometry_poni1:.18e} m"

    geometry_poni2 = flatdict.pop("geometry.poni2", None)
    if geometry_poni2 is not None:
        flatdict["center dim1"] = f"{geometry_poni2:.18e} m"

    geometry_rot1 = flatdict.pop("geometry.rot1", None)
    if geometry_rot1 is not None:
        flatdict["rot1"] = f"{geometry_rot1:.18e} rad"

    geometry_rot2 = flatdict.pop("geometry.rot2", None)
    if geometry_rot2 is not None:
        flatdict["rot2"] = f"{geometry_rot2:.18e} rad"

    geometry_rot3 = flatdict.pop("geometry.rot3", None)
    if geometry_rot3 is not None:
        flatdict["rot3"] = f"{geometry_rot3:.18e} rad"

    return [f"{k} = {v}" for k, v in flatdict.items()]


def integration_info_as_nxdict(
    integration_info: Optional[Dict[str, Any]], as_nxnote: bool = True
) -> Dict[str, Any]:
    """Convert to a Nexus dictionary. Add keys and units when appropriate."""
    configuration: Dict[str, Any] = {}
    multigeometry_first_poni = None
    if integration_info:
        configuration.update(
            {k: v for k, v in integration_info.items() if k != MULTIGEOMETRY_FIRST_PONI}
        )
        multigeometry_first_poni = integration_info.get(MULTIGEOMETRY_FIRST_PONI)

    nxtree_dict: Dict[str, Any] = {
        "@NX_class": "NXprocess",
        "program": "pyFAI",
        "version": _PYFAI_VERSION,
    }
    if as_nxnote:
        nxtree_dict["configuration"] = {
            "@NX_class": "NXnote",
            "type": "application/json",
            "data": json.dumps(configuration, cls=PyFaiEncoder),
        }
    else:
        configuration["@NX_class"] = "NXparameters"
        nxtree_dict["configuration"] = configuration
        if "energy" in configuration:
            configuration["energy@units"] = "keV"
        if "wavelength" in configuration:
            configuration["wavelength@units"] = "m"
        geometry = configuration.get("geometry", dict())
        if "dist" in geometry:
            geometry["dist@units"] = "m"
        if "poni1" in geometry:
            geometry["poni1@units"] = "m"
        if "poni2" in geometry:
            geometry["poni2@units"] = "m"
        if "rot1" in geometry:
            geometry["rot1@units"] = "rad"
        if "rot2" in geometry:
            geometry["rot2@units"] = "rad"
        if "rot3" in geometry:
            geometry["rot3@units"] = "rad"
    if multigeometry_first_poni:
        nxtree_dict[MULTIGEOMETRY_FIRST_PONI] = _multigeometry_first_ai_as_nxdict(
            multigeometry_first_poni, as_nxnote
        )

    return nxtree_dict


def _multigeometry_first_ai_as_nxdict(
    multigeometry_first_poni: dict, as_nxnote: bool
) -> dict:
    if as_nxnote:
        return {
            "@NX_class": "NXnote",
            "type": "application/json",
            "data": json.dumps(multigeometry_first_poni, cls=PyFaiEncoder),
            "@description": "PONI information from the first motor position.",
        }
    nxdict = {
        "@NX_class": "NXparameters",
        "@description": "PONI information from the first motor position.",
    }
    for key, value in multigeometry_first_poni.items():
        nxdict[key] = {"data": value}
        if key in ["poni1", "poni2", "dist"]:
            nxdict[key]["@units"] = "m"
        elif key in ["rot1", "rot2", "rot3"]:
            nxdict[key]["@units"] = "rad"
    return nxdict


class PyFaiEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, (numpy.generic, numpy.ndarray)):
            return obj.tolist()
        if isinstance(obj, DataUrl):
            return obj.path()
        return super().default(obj)


def _flatten_dict(
    adict: Mapping, _prefix: Optional[str] = None
) -> Iterator[Tuple[str, Any]]:
    if _prefix is None:
        _prefix = ""
    for k, v in adict.items():
        k = _prefix + k
        if isinstance(v, Mapping):
            yield from _flatten_dict(v, _prefix=f"{k}.")
        else:
            yield k, v


def _add_extra(adict: Mapping, extra: Mapping):
    for k, v in extra.items():
        if v is not None:
            adict[k] = v


def create_nxprocess(
    parent: h5py.Group,
    link_parent: h5py.Group,
    nxprocess_name: str,
    info: Optional[Dict[str, Any]],
    overwrite: bool = False,
) -> h5py.Group:
    """Create NXprocess group in the external parent with link in the parent when these are different groups."""
    nxtree_dict = integration_info_as_nxdict(info)
    nxprocess_path = f"{parent.name}/{nxprocess_name}"
    dicttonx(nxtree_dict, parent.file, h5path=nxprocess_path, update_mode="modify")
    nxprocess = parent[nxprocess_name]
    create_hdf5_link(link_parent, nxprocess_name, nxprocess, overwrite=overwrite)
    return nxprocess


def create_integration_results_nxdata(
    nxprocess: h5py.Group,
    intensity_dim: int,
    radial: numpy.ndarray,
    radial_units: str,
    azimuthal: Union[numpy.ndarray, None],
    azimuthal_units: str,
    overwrite: bool = False,
) -> h5py.Group:
    group_name = "integrated"
    if overwrite and group_name in nxprocess:
        del nxprocess[group_name]
    nxdata = nxprocess.create_group(group_name)
    nxdata.attrs["NX_class"] = "NXdata"
    nxprocess.attrs["default"] = group_name

    # Axes interpretation
    add_axes_to_nxdata(
        nxdata, intensity_dim, radial, radial_units, azimuthal, azimuthal_units
    )

    return nxdata


def add_axes_to_nxdata(
    nxdata: h5py.Group,
    intensity_dim: int,
    radial: numpy.ndarray,
    radial_units,
    azimuthal: Union[numpy.ndarray, None],
    azimuthal_units,
):
    # Axes names and units
    radial_units = data_from_storage(radial_units, remove_numpy=True)
    try:
        radial_axis = parse_string_units(radial_units)
    except ValueError as e:
        raise TaskInputError(e)
    has_azimuth = not is_missing_data(azimuthal) and azimuthal is not None
    if has_azimuth:
        azimuthal_axis = parse_string_units(azimuthal_units)

    if has_azimuth and intensity_dim == 2:
        nxdata.attrs["axes"] = [azimuthal_axis.name, radial_axis.name]
        nxdata.attrs["interpretation"] = "image"
    elif has_azimuth and intensity_dim == 3:
        nxdata.attrs["axes"] = [".", azimuthal_axis.name, radial_axis.name]
        nxdata.attrs["interpretation"] = "image"
    elif not has_azimuth and intensity_dim == 2:
        nxdata.attrs["axes"] = [".", radial_axis.name]
        nxdata.attrs["interpretation"] = "spectrum"
    elif not has_azimuth and intensity_dim == 1:
        nxdata.attrs["axes"] = [radial_axis.name]
        nxdata.attrs["interpretation"] = "spectrum"
    else:
        raise ValueError("Unrecognized data")

    dset = nxdata.create_dataset(radial_axis.name, data=radial)
    dset.attrs["units"] = radial_axis.units
    if has_azimuth:
        dset = nxdata.create_dataset(azimuthal_axis.name, data=azimuthal)
        dset.attrs["units"] = azimuthal_axis.units


def pyfai_to_goniometer(angles: numpy.ndarray) -> numpy.ndarray:
    """Converts an azimuthal angle in pyFAI's coodinate system (counter-clockwise, origin at horizontal, radians -pi+pi)
    and returns it as goniometer angle (orientation 3: counter-clockwise, origin at beam center, in degree 0-360)
    """
    return ((-numpy.rad2deg(angles)) - 180) % 360


def ai_to_poni(ai: AzimuthalIntegrator) -> Dict[str, any]:
    return {
        "poni1": ai._poni1,
        "poni2": ai._poni2,
        "dist": ai._dist,
        "rot1": ai._rot1,
        "rot2": ai._rot2,
        "rot3": ai._rot3,
        "detector": getattr(ai.detector, "name", None),
        "wavelength": getattr(ai, "wavelength", None),
    }



import re
import zipfile

import h5py
import numpy
import pytest
from ewoksorange.tests.utils import execute_task
from silx.io.dictdump import dicttonx

from orangecontrib.ewoksxrpd.ascii import OWSaveAsciiPattern1D

from ..tasks.ascii import SaveAsciiMultiPattern1D
from ..tasks.ascii import SaveNexusPatternsAsAscii


def test_save_ascii_task(tmpdir, setup1):
    assert_save_ascii(tmpdir, setup1, None)


def test_save_ascii_widget(tmpdir, setup1, qtapp):
    assert_save_ascii(tmpdir, setup1, qtapp)


def assert_save_ascii(tmpdir, setup1, qtapp):
    inputs = {
        "filename": str(tmpdir / "result.dat"),
        "x": numpy.linspace(1, 60, 60),
        "y": numpy.random.random(60),
        "xunits": "2th_deg",
        "header": {
            "energy": 10.2,
            "detector": setup1.detector,
            "detector_config": setup1.detector_config,
            "geometry": setup1.geometry,
        },
        "metadata": {"name": "mysample"},
    }

    execute_task(
        OWSaveAsciiPattern1D.ewokstaskclass if qtapp is None else OWSaveAsciiPattern1D,
        inputs=inputs,
    )

    x, y = numpy.loadtxt(str(tmpdir / "result.dat")).T
    numpy.testing.assert_array_equal(x, inputs["x"])
    numpy.testing.assert_array_equal(y, inputs["y"])

    with open(tmpdir / "result.dat") as f:
        lines = list()
        for line in f:
            if not line.startswith("#"):
                break
            lines.append(line)
    lines = "".join(lines)

    for key in (
        "detector",
        "energy",
        "distance",
        "center dim0",
        "center dim1",
        "rot1",
        "rot2",
        "rot3",
        "xunits",
    ):
        assert f"{key} =" in lines
    assert "name = mysample" in lines
    m = re.findall("energy = (.+) keV", lines)
    assert len(m) == 1
    assert float(m[0]) == inputs["header"]["energy"]


def test_save_multi_ascii(tmpdir, setup1):
    inputs = {
        "filenames": [str(tmpdir / "result1.dat"), str(tmpdir / "result2.dat")],
        "x_list": [numpy.linspace(1, 60, 60), numpy.linspace(1, 60, 60)],
        "y_list": [numpy.random.random(60), numpy.random.random(60)],
        "yerror_list": [numpy.random.random(60), numpy.random.random(60)],
        "xunits_list": ["2th_deg", "2th_rad"],
        "header_list": [
            {
                "energy": 10.2,
                "detector": setup1.detector,
                "geometry": setup1.geometry,
            },
            {
                "energy": 9.8,
                "detector": setup1.detector,
                "geometry": setup1.geometry,
            },
        ],
        "metadata_list": [{"name": "mysample"}, {"name": "mysample"}],
    }

    execute_task(SaveAsciiMultiPattern1D, inputs=inputs)

    for (
        filename,
        input_x,
        input_y,
        input_yerror,
        input_header,
        input_metadata,
    ) in zip(
        inputs["filenames"],
        inputs["x_list"],
        inputs["y_list"],
        inputs["yerror_list"],
        inputs["header_list"],
        inputs["metadata_list"],
    ):
        x, y, yerror = numpy.loadtxt(filename).T
        numpy.testing.assert_array_equal(x, input_x)
        numpy.testing.assert_array_equal(y, input_y)
        numpy.testing.assert_array_equal(yerror, input_yerror)

        with open(filename) as f:
            lines = list()
            for line in f:
                if not line.startswith("#"):
                    break
                lines.append(line)
        lines = "".join(lines)

        for key in (
            "detector",
            "energy",
            "distance",
            "center dim0",
            "center dim1",
            "rot1",
            "rot2",
            "rot3",
            "xunits",
        ):
            assert f"{key} =" in lines
        for k, v in input_metadata.items():
            assert f"{k} = {v}" in lines
        matches = re.findall("energy = (.+) keV", lines)
        assert len(matches) == 1
        assert float(matches[0]) == input_header["energy"]


def test_SaveNexusPatternsAsAscii_single_pattern(tmp_path):
    """Test with a 1D signal, no errors and no "points" axis"""
    input_filename = str(tmp_path / "input.h5")
    nxdata_content = {
        "@NX_class": "NXdata",
        "@axes": ["q"],
        "@interpretation": "spectrum",
        "@signal": "intensity",
        "intensity": 2 * numpy.ones(100, dtype=numpy.float32),
        "q": numpy.linspace(0.01, 10, 100, dtype=numpy.float64),
        "q@units": "A^-1",
    }
    dicttonx(nxdata_content, input_filename, "/entry/integrated")

    task = SaveNexusPatternsAsAscii(
        inputs={
            "nxdata_url": f"{input_filename}::/entry/integrated",
            "output_filename_template": str(tmp_path / "output_%04d.xye"),
        }
    )
    task.execute()

    expected_filepath = tmp_path / "output_0000.xye"
    assert task.outputs.filenames == (str(expected_filepath),)

    # Check header
    text = expected_filepath.read_text().splitlines()
    assert "# xunits = A^-1" in text, "xunit is missing or wrong"

    # Check data
    assert numpy.array_equal(
        numpy.loadtxt(expected_filepath),
        # Expected array content:
        numpy.transpose(
            [
                nxdata_content["q"],
                nxdata_content["intensity"],
            ]
        ),
    ), "Saved data differs from original data"


def test_SaveNexusPatternsAsAscii_multi_patterns(
    tmp_path, multi_integrated_pattern_file
):
    """Test with a 2D signal, no errors and no "points" axis"""

    task = SaveNexusPatternsAsAscii(
        inputs={
            "nxdata_url": f"{multi_integrated_pattern_file}::/entry/integrated",
            "output_filename_template": str(tmp_path / "output_%04d.xye"),
            "header": {"info": "test"},
        }
    )
    task.execute()

    expected_filepaths = (
        tmp_path / "output_0000.xye",
        tmp_path / "output_0001.xye",
    )
    assert task.outputs.filenames == tuple(str(p) for p in expected_filepaths)

    with h5py.File(multi_integrated_pattern_file, "r") as h5file:
        input_q = h5file["/entry/integrated/q"][()]
        input_intensity = h5file["/entry/integrated/intensity"][()]

    for index, filepath in enumerate(expected_filepaths):
        # Check header
        text = filepath.read_text().splitlines()
        assert "# point =" not in text, "No point header expected"
        assert "# xunits = A^-1" in text, "xunit is missing or wrong"
        assert "# info = test" in text, "Header information is missing"

        # Check data
        assert numpy.array_equal(
            numpy.loadtxt(filepath),
            # Expected array content:
            numpy.transpose(
                [
                    input_q,
                    input_intensity[index],
                ]
            ),
        ), "Saved data differs from original data"


def test_SaveNexusPatternsAsAscii_multi_patterns_with_errors_and_points(tmp_path):
    """Test with a 2D signal, errors and a first "points" axis"""
    input_filename = str(tmp_path / "input.h5")
    nxdata_content = {
        "@NX_class": "NXdata",
        "@axes": ["points", "q"],
        "@interpretation": "spectrum",
        "@signal": "intensity",
        "intensity": 2 * numpy.ones((2, 100), dtype=numpy.float32),
        "intensity_errors": numpy.ones((2, 100), dtype=numpy.float32),
        "q": numpy.linspace(0.01, 10, 100, dtype=numpy.float64),
        "q@units": "A^-1",
        "points": [0, 1],
    }
    dicttonx(nxdata_content, input_filename, "/entry/integrated")

    task = SaveNexusPatternsAsAscii(
        inputs={
            "nxdata_url": f"{input_filename}::/entry/integrated",
            "output_filename_template": str(tmp_path / "output_%04d.xye"),
            "header": {"info": "test"},
        }
    )
    task.execute()

    expected_filepaths = (
        tmp_path / "output_0000.xye",
        tmp_path / "output_0001.xye",
    )
    assert task.outputs.filenames == tuple(str(p) for p in expected_filepaths)

    for index, filepath in enumerate(expected_filepaths):
        # Check header
        text = filepath.read_text().splitlines()
        assert (
            f"# point = {nxdata_content['points'][index]}" in text
        ), "Corresponding point value is missing or wrong"
        assert "# xunits = A^-1" in text, "xunit is missing or wrong"
        assert "# info = test" in text, "Header information is missing"

        # Check data
        assert numpy.array_equal(
            numpy.loadtxt(filepath),
            # Expected array content:
            numpy.transpose(
                [
                    nxdata_content["q"],
                    nxdata_content["intensity"][index],
                    nxdata_content["intensity_errors"][index],
                ]
            ),
        ), "Saved data differs from original data"


def test_SaveNexusPatternsAsAscii_zip(tmp_path, multi_integrated_pattern_file):
    """Test save as a ZIP file"""

    zip_filename = str(tmp_path / "archive.zip")
    task = SaveNexusPatternsAsAscii(
        inputs={
            "nxdata_url": f"{multi_integrated_pattern_file}::/entry/integrated",
            "output_filename_template": "output_%04d.xye",
            "output_archive_filename": zip_filename,
            "header": {"info": "test"},
        }
    )
    task.execute()

    assert task.outputs.filenames == (zip_filename,)

    extracted_path = tmp_path / "zip_content"
    zipf = zipfile.ZipFile(zip_filename)
    zipf.extractall(path=extracted_path)

    expected_filenames = "output_0000.xye", "output_0001.xye"
    assert tuple(zipf.namelist()) == expected_filenames

    with h5py.File(multi_integrated_pattern_file, "r") as h5file:
        input_q = h5file["/entry/integrated/q"][()]
        input_intensity = h5file["/entry/integrated/intensity"][()]

    for index, filename in enumerate(expected_filenames):
        filepath = extracted_path / filename
        # Check header
        text = filepath.read_text().splitlines()
        assert "# point =" not in text, "No point header expected"
        assert "# xunits = A^-1" in text, "xunit is missing or wrong"
        assert "# info = test" in text, "Header information is missing"

        # Check data
        assert numpy.array_equal(
            numpy.loadtxt(filepath),
            # Expected array content:
            numpy.transpose(
                [
                    input_q,
                    input_intensity[index],
                ]
            ),
        ), "Saved data differs from original data"


def test_overwrite_SaveNexusPatternsAsAscii(tmp_path, multi_integrated_pattern_file):
    expected_filepaths = (
        tmp_path / "output_0000.xye",
        tmp_path / "output_0001.xye",
    )

    for filepath in expected_filepaths:
        filepath.touch()
        assert filepath.exists()

    inputs = {
        "nxdata_url": f"{multi_integrated_pattern_file}::/entry/integrated",
        "output_filename_template": str(tmp_path / "output_%04d.xye"),
        "header": {"info": "test"},
    }

    task_without_overwrite = SaveNexusPatternsAsAscii(inputs=inputs)
    with pytest.raises(RuntimeError) as exc:
        task_without_overwrite.execute()
        assert isinstance(exc.__cause__, FileExistsError)

    task = SaveNexusPatternsAsAscii(
        inputs={
            **inputs,
            "overwrite": True,
        }
    )
    task.execute()

    assert task.outputs.filenames == tuple(str(p) for p in expected_filepaths)


def test_overwrite_zip_SaveNexusPatternsAsAscii(
    tmp_path, multi_integrated_pattern_file
):
    zip_filename = tmp_path / "archive.zip"
    zip_filename.touch()
    assert zip_filename.exists()

    inputs = {
        "nxdata_url": f"{multi_integrated_pattern_file}::/entry/integrated",
        "output_filename_template": "output_%04d.xye",
        "output_archive_filename": zip_filename,
        "header": {"info": "test"},
    }
    task_without_overwrite = SaveNexusPatternsAsAscii(inputs=inputs)
    with pytest.raises(RuntimeError) as exc:
        task_without_overwrite.execute()
        assert isinstance(exc.__cause__, FileExistsError)

    task = SaveNexusPatternsAsAscii(
        inputs={
            **inputs,
            "overwrite": True,
        }
    )
    task.execute()

    assert task.outputs.filenames == (zip_filename,)

    extracted_path = tmp_path / "zip_content"
    zipf = zipfile.ZipFile(zip_filename)
    zipf.extractall(path=extracted_path)

    expected_filenames = "output_0000.xye", "output_0001.xye"
    assert tuple(zipf.namelist()) == expected_filenames



import json
from pathlib import Path
from typing import Any
from typing import Dict

import h5py
import numpy
import pytest
from silx.io.url import DataUrl

from ..tasks.integrate import IntegrateBlissScan
from ..tasks.multi_geometry_integrate import Integrate2DMultiGeometry
from ..tasks.utils import pyfai_utils
from . import xrpd_theory


@pytest.fixture
def gonio_file(tmp_path: Path) -> Path:
    gonio_json = {
        "content": "Goniometer calibration v2",
        "detector": "Pilatus1M",
        "wavelength": 2.6379616687914948e-11,
        "param": [
            2.0062960356379422,
            0.015883422992655393,
            0.8216006393975213,
            0.002164173240804926,
            -0.0031180017089310853,
            0.0063336570464145835,
            0.004297460651999008,
        ],
        "param_names": ["dist", "poni1", "poni2", "rot1", "rot2", "rot_x", "rot_y"],
        "pos_names": ["pos"],
        "trans_function": {
            "content": "GeometryTransformation",
            "param_names": ["dist", "poni1", "poni2", "rot1", "rot2", "rot_x", "rot_y"],
            "pos_names": ["pos"],
            "dist_expr": "dist",
            "poni1_expr": "poni1 + rot_x*cos(pos) - rot_y*sin(pos)",
            "poni2_expr": "poni2 + rot_x*sin(pos) + rot_y*cos(pos)",
            "rot1_expr": "+rot1*cos(pos) - rot2*sin(pos)",
            "rot2_expr": "+rot1*sin(pos) + rot2*cos(pos)",
            "rot3_expr": "pos",
            "constants": {"pi": 3.141592653589793},
        },
    }
    filename = tmp_path / "gonio.json"
    with open(filename, "w") as f:
        json.dump(gonio_json, f)

    return filename


@pytest.fixture
def image_data_url(
    imageSetup1Calibrant1: xrpd_theory.Calibration, tmp_path: Path
) -> DataUrl:
    assert isinstance(imageSetup1Calibrant1, xrpd_theory.Calibration)
    images = numpy.zeros(shape=(10, *imageSetup1Calibrant1.image.shape))
    for i in range(10):
        images[i] = imageSetup1Calibrant1.image
    h5_path = tmp_path / "output.h5"
    dataset_path = "/entry/data"
    with h5py.File(h5_path, "w") as h5file:
        h5file.create_dataset(dataset_path, data=images)
    return DataUrl(file_path=str(h5_path), data_path=dataset_path)


def test_multi_geometry_task(image_data_url: DataUrl, gonio_file: Path):
    task = Integrate2DMultiGeometry(
        inputs={
            "goniometer_file": gonio_file,
            "integration_options": {"npt_rad": 300, "npt_azim": 180},
            "images": image_data_url.path(),
            "positions": [*numpy.linspace(0, 360, 10)],
        }
    )

    task.execute()

    assert task.get_output_value("radial").shape == (300,)
    assert task.get_output_value("azimuthal").shape == (180,)
    assert task.get_output_value("intensity").shape == (180, 300)


def test_integration_options_consistency(
    bliss_task_inputs: Dict[str, Any],
    image_data_url: DataUrl,
    gonio_file: Path,
    tmp_path: Path,
):
    N_2TH = 300
    N_CHI = 180
    RADIAL_MIN = 0
    RADIAL_MAX = 50
    integration_options = {
        "nbpt_rad": N_2TH,
        "nbpt_azim": N_CHI,
        "radial_range_min": 0,
        "radial_range_max": 50,
    }

    mg_task = Integrate2DMultiGeometry(
        inputs={
            "goniometer_file": gonio_file,
            "integration_options": integration_options,
            "images": image_data_url.path(),
            "positions": [*numpy.linspace(0, 360, 10)],
        }
    )

    mg_task.execute()

    radial = mg_task.get_output_value("radial")
    assert radial.shape == (N_2TH,)
    assert radial.min() > RADIAL_MIN
    assert radial.max() < RADIAL_MAX
    assert mg_task.get_output_value("azimuthal").shape == (N_CHI,)
    assert mg_task.get_output_value("intensity").shape == (N_CHI, N_2TH)

    sg_task = IntegrateBlissScan(
        inputs={
            **bliss_task_inputs,
            "integration_options": {
                **integration_options,
                "integrator_name": "integrate2d_ng",
            },
            "output_filename": tmp_path / "output.h5",
        }
    )

    sg_task.execute()

    with h5py.File(tmp_path / "output.h5", "r") as h5file:
        n_frames = len(h5file["/2.1/measurement/p3"])
        integrated_grp = h5file["/2.1/p3_integrate/integrated"]
        assert isinstance(integrated_grp, h5py.Group)

        radial = integrated_grp["2th"][()]
        assert radial.shape == (N_2TH,)
        assert radial.min() > RADIAL_MIN
        assert radial.max() < RADIAL_MAX
        assert integrated_grp["chi"].shape == (N_CHI,)
        assert integrated_grp["intensity"].shape == (n_frames, N_CHI, N_2TH)


def test_integration_info_content(image_data_url: DataUrl, gonio_file: Path):
    task = Integrate2DMultiGeometry(
        inputs={
            "goniometer_file": gonio_file,
            "integration_options": {"nbpt_rad": 5000, "nbpt_azim": 360},
            "images": image_data_url.path(),
            "positions": list(numpy.linspace(0, 360, 10)),
        }
    )

    task.execute()
    info = task.get_output_value("info")

    assert "nbpt_rad" in info
    assert "nbpt_azim" in info
    assert "goniometer" in info

    goniometer_info = info["goniometer"]
    expected_keys = [
        "content",
        "detector",
        "wavelength",
        "param",
        "param_names",
        "pos_names",
        "trans_function",
    ]
    for key in expected_keys:
        assert key in goniometer_info

    assert len(goniometer_info["param"]) == len(goniometer_info["param_names"])

    assert pyfai_utils.MULTIGEOMETRY_FIRST_PONI in info
    first_ai = info[pyfai_utils.MULTIGEOMETRY_FIRST_PONI]
    for key in ["poni1", "poni2", "dist", "rot1", "rot2", "rot3"]:
        assert key in first_ai

    assert json.dumps(info)



import json

import pytest
from ewokscore import execute_graph
from ewoksorange.tests.utils import execute_task
from pyFAI.io.ponifile import PoniFile

from orangecontrib.ewoksxrpd.savepyfaiconfig import OWSavePyFaiConfig
from orangecontrib.ewoksxrpd.savepyfaiponifile import OWSavePyFaiPoniFile

from ..tasks.pyfaiconfig import PyFaiConfig
from ..tasks.pyfaiconfig import SavePyFaiConfig
from ..tasks.pyfaiconfig import SavePyFaiPoniFile


def test_pyfai_config_roundtrip(tmp_path, setup1):
    """Create and save pyFAI configuration"""
    workflow = {
        "graph": {"id": "test_roundtrip"},
        "nodes": [
            {
                "id": "load",
                "task_type": "class",
                "task_identifier": "ewoksxrpd.tasks.pyfaiconfig.PyFaiConfig",
            },
            {
                "id": "save",
                "task_type": "class",
                "task_identifier": "ewoksxrpd.tasks.pyfaiconfig.SavePyFaiConfig",
            },
        ],
        "links": [
            {
                "source": "load",
                "target": "save",
                "data_mapping": [
                    {"source_output": "energy", "target_input": "energy"},
                    {"source_output": "geometry", "target_input": "geometry"},
                    {"source_output": "detector", "target_input": "detector"},
                    {
                        "source_output": "detector_config",
                        "target_input": "detector_config",
                    },
                    {"source_output": "mask", "target_input": "mask"},
                    {
                        "source_output": "integration_options",
                        "target_input": "integration_options",
                    },
                ],
            },
        ],
    }

    output_path = tmp_path / "pyfaiconfig.json"
    mask_filename = str(tmp_path / "mask.edf")
    integration_options = {"error_model": "poisson", **setup1.integration_options}
    result = execute_graph(
        workflow,
        inputs=[
            {"id": "load", "name": "energy", "value": setup1.energy},
            {"id": "load", "name": "geometry", "value": setup1.geometry},
            {"id": "load", "name": "mask", "value": mask_filename},
            {"id": "load", "name": "detector", "value": setup1.detector},
            {"id": "load", "name": "detector_config", "value": setup1.detector_config},
            {"id": "load", "name": "integration_options", "value": integration_options},
            {"id": "save", "name": "output_filename", "value": str(output_path)},
        ],
        outputs=[{"all": False}],
    )
    assert result["filename"] == str(output_path)

    version = setup1.version or 3
    poni = {
        "wavelength": setup1.wavelength,
        "detector": setup1.detector,
        "detector_config": setup1.load_detector_config(),
        **setup1.geometry,
    }
    expected_config = {
        "application": "pyfai-integrate",
        "version": version,
        "do_mask": True,
        "mask_file": mask_filename,
        **integration_options,
    }
    if version >= 4:
        expected_config["poni"] = poni
    else:
        expected_config.update(poni)

    config = json.loads(output_path.read_text())
    assert config == expected_config


def test_SavePyFaiConfig(tmp_path, setup1, qtapp):
    output_path = tmp_path / "pyfaiconfig.json"
    integration_options = {"error_model": "poisson", **setup1.integration_options}
    mask_filename = str(tmp_path / "mask.edf")
    inputs = {
        "output_filename": str(output_path),
        "energy": setup1.energy,
        "geometry": setup1.geometry,
        "detector": setup1.detector,
        "mask": mask_filename,
        "detector_config": setup1.detector_config,
        "integration_options": integration_options,
    }

    result = execute_task(
        SavePyFaiConfig if qtapp is None else OWSavePyFaiConfig,
        inputs=inputs,
    )

    assert result["filename"] == str(output_path)

    version = setup1.version or 3
    poni = {
        "wavelength": setup1.wavelength,
        "detector": setup1.detector,
        "detector_config": setup1.load_detector_config(),
        **setup1.geometry,
    }
    expected_config = {
        "application": "pyfai-integrate",
        "version": version,
        "do_mask": True,
        "mask_file": mask_filename,
        **integration_options,
    }
    if version >= 4:
        expected_config["poni"] = poni
    else:
        expected_config.update(poni)

    config = json.loads(output_path.read_text())
    assert config == expected_config


def test_SavePyFaiConfig_v4(tmp_path, setup1, qtapp):
    output_path = tmp_path / "pyfaiconfig.json"

    bad_poni_info = {
        **setup1.geometry,
        "wavelength": 0.0,
        "detector": "BAD_DETECTOR",
        "detector_config": {"orientation": 1},
    }
    integration_options = {
        "application": "pyfai-integrate",
        "version": 4,
        "error_model": "poisson",
        "poni": bad_poni_info,
    }

    mask_filename = str(tmp_path / "mask.edf")
    inputs = {
        "output_filename": str(output_path),
        "energy": setup1.energy,
        "geometry": setup1.geometry,
        "detector": setup1.detector,
        "mask": mask_filename,
        "detector_config": setup1.detector_config,
        "integration_options": integration_options,
    }

    result = execute_task(
        SavePyFaiConfig if qtapp is None else OWSavePyFaiConfig,
        inputs=inputs,
    )

    assert result["filename"] == str(output_path)
    expected_config = {
        "application": "pyfai-integrate",
        "version": 4,
        "poni": {
            **setup1.geometry,
            "wavelength": setup1.wavelength,
            "detector": setup1.detector,
            "detector_config": setup1.load_detector_config(),
        },
        "do_mask": True,
        "mask_file": mask_filename,
        "error_model": "poisson",
    }
    config = json.loads(output_path.read_text())
    assert config == expected_config


def test_SavePyFaiPoniFile(tmp_path, setup1, qtapp):
    output_path = tmp_path / "pyfaiconfig.json"

    inputs = {
        "output_filename": str(output_path),
        "energy": setup1.energy,
        "geometry": setup1.geometry,
        "detector": setup1.detector,
        "detector_config": setup1.detector_config,
        "integration_options": setup1.integration_options,
    }

    result = execute_task(
        SavePyFaiPoniFile if qtapp is None else OWSavePyFaiPoniFile,
        inputs=inputs,
    )

    assert result["filename"] == str(output_path)

    result_poni = PoniFile(result["filename"])
    expected_poni = PoniFile(
        {
            "wavelength": setup1.wavelength,
            **setup1.geometry,
            "detector": setup1.detector,
            "detector_config": setup1.detector_config,
        }
    )
    assert result_poni.as_dict() == expected_poni.as_dict()


@pytest.mark.parametrize("show_warnings", (True, False))
def test_merging_of_config(tmp_path, caplog, show_warnings):
    energy = 12.5
    filename = tmp_path / "pyfaiconfig.json"
    filename2 = tmp_path / "pyfaiconfig2.json"

    with open(filename, "w") as f:
        json.dump(
            {
                "npt_rad": 100,
                "energy": energy,
            },
            f,
        )

    with open(filename2, "w") as f:
        json.dump(
            {"npt_rad": 400},
            f,
        )

    inputs = {
        "filename": str(filename),
        "filenames": [str(filename2)],
        "integration_options": {"error_model": "poisson"},
        "show_merge_warnings": show_warnings,
    }

    outputs = execute_task(PyFaiConfig, inputs=inputs)

    if show_warnings:
        assert (
            f"New value of 'npt_rad' (400) from {str(filename2)} will overwrite the current value (100)"
            in caplog.text
        )
    else:
        assert caplog.text == ""

    assert outputs["energy"] == energy
    assert outputs["integration_options"] == {"npt_rad": 400, "error_model": "poisson"}



import os
import warnings
import zipfile
from pathlib import Path

from ..tasks.utils.zip_utils import open_in_zipfile


def test_extracted_file_contents(tmp_path: Path):
    archive_path = tmp_path / "archive.zip"
    extracted_path = tmp_path / "zip_content"
    file_inside_zip = "folder/nested.txt"
    file_contents = b"zip test!"

    with zipfile.ZipFile(archive_path, mode="w") as zipf:
        with open_in_zipfile(zipf, file_inside_zip, mode="w") as f:
            f.write(file_contents)

    _unzip(archive_path, extracted_path)

    extracted_file = extracted_path / file_inside_zip

    assert extracted_file.exists()
    assert extracted_file.read_bytes() == file_contents


def test_permissions_applied(tmp_path: Path):
    archive_path = tmp_path / "archive.zip"
    extracted_path = tmp_path / "unzipped"
    file_inside_zip = "data.txt"

    with zipfile.ZipFile(archive_path, mode="w") as zipf:
        with open_in_zipfile(zipf, file_inside_zip, mode="w") as f:
            f.write(b"abc")

    _unzip(archive_path, extracted_path)

    extracted_file = extracted_path / file_inside_zip
    assert extracted_file.exists()

    with zipfile.ZipFile(archive_path) as zipf:
        info = zipf.getinfo(file_inside_zip)
        perm_bits = (info.external_attr >> 16) & 0o777

    current_umask = os.umask(0)
    os.umask(current_umask)
    expected = 0o666 & ~current_umask

    assert perm_bits == expected


def test_pathlib_support(tmp_path: Path):
    archive_path = tmp_path / "archive.zip"
    extracted_path = tmp_path / "unzipped"
    internal_file = Path("subdir") / "file.txt"
    data = b"12345"

    with zipfile.ZipFile(archive_path, mode="w") as zipf:
        with open_in_zipfile(zipf, internal_file, mode="w") as f:
            f.write(data)

    _unzip(archive_path, extracted_path)

    extracted_file = extracted_path / internal_file
    assert extracted_file.exists()
    assert extracted_file.read_bytes() == data


def test_multiple_files_in_one_archive(tmp_path: Path):
    archive_path = tmp_path / "archive.zip"
    extracted_path = tmp_path / "unzipped"
    contents = (
        ("a.txt", b"AAA"),
        ("b/b.txt", b"BBB1"),
        ("b/b.txt", b"BBB2"),
        ("c/c/c.txt", b"CCC"),
    )
    final_content = {
        "a.txt": b"AAA",
        "b/b.txt": b"BBB2",
        "c/c/c.txt": b"CCC",
    }

    with zipfile.ZipFile(archive_path, mode="w") as zipf:
        for name, content in contents:
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", message="Duplicate name:.*")
                with open_in_zipfile(zipf, name, mode="w") as f:
                    f.write(content)

    _unzip(archive_path, extracted_path)

    for name, expected in final_content.items():
        extracted_file = extracted_path / name
        assert extracted_file.exists()
        assert extracted_file.read_bytes() == expected


def _unzip(archive_path: Path, extracted_path: Path) -> None:
    with zipfile.ZipFile(archive_path) as zipf:
        zipf.extractall(path=extracted_path)



import importlib.metadata
import logging
from importlib.metadata import version

import h5py
import numpy
from ewoksdata.data.nexus import select_default_plot
from packaging.version import Version
from silx.io.url import DataUrl

from .data_access import TaskWithDataAccess
from .utils import pyfai_utils
from .utils.data_utils import create_hdf5_link
from .utils.data_utils import is_same_file
from .utils.data_utils import split_hdf5_url_parent_data_path
from .utils.nexus_utils import IntegratedPattern
from .utils.nexus_utils import read_nexus_integrated_patterns

_logger = logging.getLogger(__name__)

SILX_VERSION = Version(version("silx"))


def is_silx_compatible() -> bool:
    # With older silx version, opening `external_nxprocess_url` raises the following error:
    # RuntimeError: Unable to synchronously open file (file locking 'ignore disabled locks' flag values don't match)
    return SILX_VERSION >= Version("2.2.0")


def _can_substract_background(
    pattern: IntegratedPattern, background_pattern: IntegratedPattern
) -> bool:
    if pattern.radial_units != background_pattern.radial_units:
        return False

    return pattern.intensity.shape == background_pattern.intensity.shape


class SubtractBackgroundPattern(
    TaskWithDataAccess,
    input_names=["nxdata_url", "background_nxdata_url", "output_nxprocess_url"],
    optional_input_names=["enabled", "background_factor", "external_nxprocess_url"],
    output_names=["nxdata_url"],
):
    """
    Subtract an integrated pattern from patterns stored in NeXus and saves the result in a new NeXus process group

    .. code:

        I_pattern_corrected = I_pattern - (I_background_pattern * background_factor)

    If errors are available in the background and in the input data, errors are propagated and saved as well.

    .. code:

        E_I_pattern_corrected = \\sqrt{E_I_pattern^2 + (E_I_background_pattern * background_factor)^2}
    """

    def run(self):
        if not is_silx_compatible():
            raise ValueError(
                f"Silx version must be above or equal 2.2.0 to use this task. The current version is {SILX_VERSION}."
            )

        if not self.get_input_value("enabled", True):
            _logger.info(
                f"Task {self.__class__.__qualname__} is disabled: no pattern was subtracted"
            )
            self.outputs.nxdata_url = self.inputs.nxdata_url
            return

        if not isinstance(self.inputs.background_nxdata_url, str):
            raise ValueError(
                f"background_nxdata_url should be a str. Got {type(self.inputs.background_nxdata_url)} instead."
            )

        background_nxdata_url = DataUrl(self.inputs.background_nxdata_url)

        with self.open_h5item(background_nxdata_url, mode="r") as background_nxdata:
            if not isinstance(background_nxdata, h5py.Group):
                raise TypeError(
                    f"{background_nxdata_url.path()} should point towards a NXData Group."
                )
            background_patterns = list(
                read_nexus_integrated_patterns(background_nxdata)
            )
            background_pattern = background_patterns[0]

        nxdata_url = DataUrl(self.inputs.nxdata_url)
        output_nxprocess_url = DataUrl(self.inputs.output_nxprocess_url)
        external_nxprocess_url = DataUrl(
            self.get_input_value(
                "external_nxprocess_url", self.inputs.output_nxprocess_url
            )
        )
        output_nxprocess_parent_url, output_nxprocess_name = (
            split_hdf5_url_parent_data_path(output_nxprocess_url)
        )

        mode = (
            "a"
            if is_same_file(nxdata_url.file_path(), external_nxprocess_url.file_path())
            else "r"
        )

        with self.open_h5item(self.inputs.nxdata_url, mode=mode) as nxdata:
            if not isinstance(nxdata, h5py.Group):
                raise TypeError(
                    f"{self.inputs.nxdata_url} should point towards a NXData Group."
                )

            background_factor = float(self.get_input_value("background_factor", 1))
            background_intensity = background_pattern.intensity * background_factor
            if background_pattern.intensity_errors is None:
                background_error = None
            else:
                background_error = (
                    background_pattern.intensity_errors * background_factor
                )

            subtracted_intensity = numpy.empty(
                nxdata["intensity"].shape, dtype=nxdata["intensity"].dtype
            )
            if "intensity_errors" in nxdata and background_error is not None:
                substracted_intensity_errors = numpy.empty(
                    nxdata["intensity_errors"].shape,
                    dtype=nxdata["intensity_errors"].dtype,
                )
            else:
                substracted_intensity_errors = None

            pattern = None
            for i, pattern in enumerate(read_nexus_integrated_patterns(nxdata)):
                if not _can_substract_background(pattern, background_pattern):
                    raise ValueError(
                        f"Background pattern {background_pattern} is not compatible with data pattern {pattern}."
                    )

                subtracted_intensity[i] = pattern.intensity - background_intensity
                if (
                    background_error is not None
                    and substracted_intensity_errors is not None
                    and pattern.intensity_errors is not None
                ):
                    substracted_intensity_errors[i] = numpy.sqrt(
                        pattern.intensity_errors**2 + background_error**2
                    )

            if pattern is None:
                raise ValueError(f"No integrated patterns in {nxdata.name}")

            with self.open_h5item(
                external_nxprocess_url, create=True, mode="a"
            ) as external_process:
                external_process["program"] = "ewoksxrpd"
                external_process["version"] = importlib.metadata.version("ewoksxrpd")

                config_group = external_process.create_group("configuration")
                config_group.attrs["NX_class"] = "NXparameters"
                for name, value in self.get_input_values().items():
                    if value is not None:
                        config_group[name] = value

                output_nxdata = pyfai_utils.create_integration_results_nxdata(
                    external_process,
                    subtracted_intensity.ndim,
                    pattern.radial,
                    f"{pattern.radial_name}_{pattern.radial_units}",
                    azimuthal=None,
                    azimuthal_units="",
                )
                if "points" in nxdata:
                    create_hdf5_link(output_nxdata, "points", nxdata["points"])
                    axes = output_nxdata.attrs["axes"]
                    output_nxdata.attrs["axes"] = ["points", *axes[1:]]
                output_nxdata.create_dataset("intensity", data=subtracted_intensity)
                output_nxdata.attrs["signal"] = "intensity"
                if substracted_intensity_errors is not None:
                    output_nxdata.create_dataset(
                        "intensity_errors", data=substracted_intensity_errors
                    )
                output_nxdata.create_dataset("background", data=background_intensity)

                select_default_plot(output_nxdata)

                if output_nxprocess_url != external_nxprocess_url:
                    with self.open_h5item(
                        output_nxprocess_parent_url, create=True, mode="a"
                    ) as output_nxprocess_parent:
                        create_hdf5_link(
                            output_nxprocess_parent,
                            output_nxprocess_name,
                            external_process,
                        )

                self.outputs.nxdata_url = (
                    f"{output_nxdata.file.filename}::{output_nxdata.name}"
                )



from __future__ import annotations

import io
import os
from typing import Any
from typing import Dict
from typing import Mapping
from typing import Union

import numpy

from .data_utils import is_data
from .pyfai_utils import integration_info_as_text


def ensure_parent_folder(filename: str):
    dirname = os.path.dirname(filename)
    if dirname:
        os.makedirs(dirname, exist_ok=True)


def save_pattern_as_ascii(
    filename: str | io.TextIOBase,
    x: numpy.ndarray,
    y: numpy.ndarray,
    xunits: str,
    yerror: Union[numpy.ndarray, None],
    header: Mapping,
    metadata: Dict[str, Any],
) -> None:
    if is_data(yerror):
        data = [x, y, yerror]
        columns = ["x", "intensity", "intensity_error"]
    else:
        data = [x, y]
        columns = ["x", "intensity"]
    data = numpy.stack(data, axis=1)

    lines = integration_info_as_text(header, xunits=xunits, **metadata)
    lines.append(" ".join(columns))

    if isinstance(filename, str):
        ensure_parent_folder(filename)
    numpy.savetxt(filename, data, header="\n".join(lines))



import logging
import time
from contextlib import contextmanager
from typing import Any
from typing import Generator
from typing import Iterator
from typing import Optional
from typing import Tuple

from ewoksdata.data.contextiterator import contextiterator

_logger = logging.getLogger(__name__)


@contextiterator
def zip_with_progress(
    *iterators: Iterator[Iterator[Any]],
    message: str = "Progress %d/%s",
    nmax: Optional[int] = None,
    period: float = 5,
    logger=None,
    iter_time_name: str = "iteration",
) -> Generator[Tuple[Any], None, None]:
    """Like python's zip but will progress logging when iterating over the result.
    In addition it yields a context manager which allows for timing different sections
    in the context as well as the yielding time itself:

    .. code-block::python

        for v1, v2, tmcontext in zip_with_progress([...], [...], iter_time_name="read"):
            with tmcontext("process"):
                ...
            with tmcontext("write"):
                ...
    """
    ctimer = _ContextTimer(iter_time_name)
    if logger is None:
        logger = _logger
    i = 0
    t0 = time.time()
    if nmax is None:
        nmax = "?"
    try:
        for tpl in zip(*iterators):
            yield (*tpl, ctimer.tmcontext)
            i += 1
            if (time.time() - t0) > period:
                t0 = time.time()
                _logger.info(f"{message} (ONGOING: {ctimer})", i, nmax)
    finally:
        _logger.info(f"{message} (FINISHED: {ctimer})", i, nmax)


class _ContextTimer:
    def __init__(self, rest_context: str = "other") -> None:
        self._total_t0 = time.perf_counter_ns()
        self._context_time = dict()
        self._rest_context = rest_context

    @contextmanager
    def tmcontext(self, name: str) -> Generator[None, None, None]:
        context_t0 = time.perf_counter_ns()
        try:
            yield
        finally:
            context_t1 = time.perf_counter_ns()
            self._context_time.setdefault(name, 0)
            self._context_time[name] += context_t1 - context_t0

    def __str__(self) -> str:
        dt_total = time.perf_counter_ns() - self._total_t0
        dt_sum = sum([dt for dt in self._context_time.values()])
        dt_rest = dt_total - dt_sum
        dts = [
            f"{name}={self.strftime(dt, dt_total)}"
            for name, dt in self._context_time.items()
        ]
        dts.append(f"{self._rest_context}={self.strftime(dt_rest, dt_total)}")
        dts.append(f"TOTAL={self.strftime(dt_total, dt_total)}")
        return ", ".join(dts)

    @staticmethod
    def strftime(dt_ns: int, dt_total: int) -> str:
        m, s = divmod(int(dt_ns * 1e-9 + 0.5), 60)
        h, m = divmod(m, 60)
        f = int(dt_ns / dt_total * 100 + 0.5)
        return f"{h:02d}:{m:02d}:{s:02d} ({f:d}%%)"



import enum

import numpy
import pytest
from pyFAI.detectors.orientation import Orientation

from ..tasks.utils import pyfai_utils


def test_builtin_types():
    parameters = {
        "key1": 10,
        "key2": "normal",
        "key3": [
            "/gpfs/jazzy",
            "/gpfs/jazzy/",
            "silx:///gpfs/jazzy/path/to/file",
            "/mnt/multipath-shares/path/to/file",
        ],
    }
    parameters["key4"] = {
        "key1": 10,
        "key2": "normal",
        "key3": [
            "/gpfs/jazzy",
            "/gpfs/jazzy/",
            "silx:///gpfs/jazzy/path/to/file",
            "/mnt/multipath-shares/path/to/file",
        ],
    }
    normalized = {
        "key1": 10,
        "key2": "normal",
        "key3": ["/gpfs/jazzy", "/", "silx:///path/to/file", "/path/to/file"],
    }
    normalized["key4"] = {
        "key1": 10,
        "key2": "normal",
        "key3": ["/gpfs/jazzy", "/", "silx:///path/to/file", "/path/to/file"],
    }
    assert pyfai_utils.normalize_parameters(parameters) == normalized


def test_enum_types():
    Color = enum.Enum("Color", ["RED", "GREEN", "BLUE"])

    parameters = {"key1": Orientation.BottomRight, "key2": Color.GREEN}
    normalized = {"key1": Orientation.BottomRight.value, "key2": 2}

    assert pyfai_utils.normalize_parameters(parameters) == normalized


def test_unexpected_type():
    class UnexpectedClass:
        pass

    with pytest.warns(
        UserWarning,
        match="Unexpected pyFAI configuration parameter type",
    ):
        parameters = {"key": UnexpectedClass()}
        _ = pyfai_utils.normalize_parameters(parameters)


@pytest.mark.parametrize(
    "input_rad, expected_deg",
    [
        (0, 180),  # 0 rad => 180°
        (numpy.pi / 2, 90),  # 90° CCW => 90°
        (-numpy.pi / 2, 270),  # -90° CW => 270°
        (numpy.pi, 0),  # PI rad => 0°
        (-numpy.pi, 0),  # PI rad => 0°
    ],
)
def test_pyfai_to_goniometer(input_rad, expected_deg):
    result = pyfai_utils.pyfai_to_goniometer(numpy.array([input_rad]))[0]
    assert numpy.isclose(
        result, expected_deg, atol=1e-10
    ), f"Expected {expected_deg}, got {result}"



from typing import Any
from typing import Dict
from typing import Tuple

from ewoksdata.data.bliss import get_image
from pyFAI.goniometer import Goniometer
from pyFAI.goniometer import GoniometerRefinement

try:
    from pyFAI.integrator.azimuthal import AzimuthalIntegrator
except ImportError:
    from pyFAI import AzimuthalIntegrator
from pyFAI.units import Unit

from .base_integrate import BaseIntegrate
from .utils import integrate_utils
from .utils import pyfai_utils


class Integrate2DMultiGeometry(
    BaseIntegrate,
    input_names=["goniometer_file", "positions", "images"],
    output_names=[
        "info",
        "radial",
        "azimuthal",
        "intensity",
        "intensity_error",
        "radial_units",
        "azimuthal_units",
    ],
):
    def run(self):
        positions: list[float] = self.inputs.positions
        raw_integration_options = self._get_ewoks_pyfai_options()
        multi_geometry_options, integrate2d_options = (
            pyfai_utils.split_multi_geom_and_integration_options(
                raw_integration_options
            )
        )
        retry_timeout = self.get_input_value("retry_timeout", None)
        retry_period = self.get_input_value("retry_period", None)

        goniometer = Goniometer.sload(self.inputs.goniometer_file)
        mg_ai = goniometer.get_mg(positions, **multi_geometry_options)
        images = get_image(
            self.inputs.images, retry_timeout=retry_timeout, retry_period=retry_period
        )
        result = mg_ai.integrate2d(images, **integrate2d_options)

        first_position = pyfai_utils.pyfai_to_goniometer(positions[0])
        first_position_ai = goniometer.get_ai(first_position)

        self.outputs.radial = result.radial
        self.outputs.azimuthal = result.azimuthal
        self.outputs.intensity = result.intensity
        self.outputs.intensity_error = integrate_utils.get_intensity_error(result)

        result_unit: Tuple[Unit, Unit] = result.unit
        radial_unit, azim_unit = result_unit
        self.outputs.radial_units = radial_unit.name
        self.outputs.azimuthal_units = azim_unit.name

        self.outputs.info = self._build_integration_info(
            raw_integration_options, goniometer, first_position_ai
        )

    def _build_integration_info(
        self,
        raw_integration_options,
        goniometer: GoniometerRefinement,
        first_position_ai: AzimuthalIntegrator,
    ) -> Dict[str, Any]:
        info = pyfai_utils.compile_integration_info(raw_integration_options)
        info["goniometer"] = goniometer.to_dict()
        info[pyfai_utils.MULTIGEOMETRY_FIRST_PONI] = pyfai_utils.ai_to_poni(
            first_position_ai
        )
        return info



import io
import os
import zipfile
from contextlib import contextmanager
from datetime import datetime
from pathlib import Path
from typing import Generator
from typing import Union


@contextmanager
def open_in_zipfile(
    zipf: zipfile.ZipFile, filename: Union[str, Path], **kwargs
) -> Generator[io.BufferedIOBase, None, None]:
    """Same as `zipfile.ZipFile.open` but with a fixed `name` argument that
    ensures the file timestamp and permissions would be equivalent as with
    python's `open` function when creating a file.
    """
    filename = str(filename).replace("\\", "/")

    info = zipfile.ZipInfo(filename)
    info.date_time = datetime.now().timetuple()[:6]

    current_umask = os.umask(0)
    os.umask(current_umask)
    file_permissions = 0o666 & ~current_umask
    info.external_attr = file_permissions << 16

    with zipf.open(info, **kwargs) as zipio:
        yield zipio



import h5py
import numpy
import pytest
from silx.io.dictdump import dicttonx
from silx.io.url import DataUrl

from ewoksxrpd.tasks.subtract_pattern import SubtractBackgroundPattern
from ewoksxrpd.tasks.subtract_pattern import is_silx_compatible

NXDATA_METADATA = {
    "@NX_class": "NXdata",
    "@axes": [".", "q"],
    "@interpretation": "spectrum",
    "@signal": "intensity",
}


@pytest.mark.skipif(not is_silx_compatible(), reason="Silx is too old")
@pytest.mark.parametrize("background_factor", (1, 0.1))
def test_execution(tmp_path, background_factor):
    q = numpy.linspace(0.01, 10, 100, dtype=numpy.float64)
    input_data = 2 * numpy.ones((2, 100), dtype=numpy.float32)
    background_data = numpy.ones((1, 100), dtype=numpy.float32)

    # Create input file
    input_filename = tmp_path / "input.h5"
    input_nxdata = {
        **NXDATA_METADATA,
        "q": q,
        "q@units": "A^-1",
        "intensity": input_data,
        "points": [0, 1],
    }
    dicttonx(input_nxdata, input_filename, "/entry/process/integrated")

    # Create background file
    bg_filename = tmp_path / "background.h5"
    bg_nxdata = {
        **NXDATA_METADATA,
        "q": q,
        "q@units": "A^-1",
        "intensity": background_data,
    }
    dicttonx(bg_nxdata, bg_filename, "/entry/process/integrated")

    inputs = {
        "nxdata_url": f"{input_filename}::/entry/process/integrated",
        "output_nxprocess_url": f"{input_filename}::/entry/process_subtracted",
        "background_nxdata_url": f"{bg_filename}::/entry/process/integrated",
        "background_factor": background_factor,
    }
    task = SubtractBackgroundPattern(inputs={**inputs})
    task.execute()

    assert (
        task.get_output_value("nxdata_url")
        == f"{input_filename}::/entry/process_subtracted/integrated"
    )

    with h5py.File(input_filename, "r") as h5file:
        input_nxdata = h5file["/entry/process/integrated"]
        output_nxdata = h5file["/entry/process_subtracted/integrated"]
        numpy.testing.assert_allclose(
            output_nxdata["intensity"][()],
            numpy.ones((2, 100), dtype=numpy.float32) * (2 - background_factor),
        )
        numpy.testing.assert_allclose(input_nxdata["q"][()], output_nxdata["q"][()])
        numpy.testing.assert_allclose(
            input_nxdata["points"][()], output_nxdata["points"][()]
        )
        for name, value in inputs.items():
            stored_value = h5file["/entry/process_subtracted/configuration"][name][()]
            if isinstance(stored_value, bytes):
                stored_value = stored_value.decode()
            assert stored_value == value


@pytest.mark.skipif(not is_silx_compatible(), reason="Silx is too old")
def test_execution_with_different_output_files(tmp_path):
    q = numpy.linspace(0.01, 10, 100, dtype=numpy.float64)
    input_data = 2 * numpy.ones((2, 100), dtype=numpy.float32)
    background_data = numpy.ones((1, 100), dtype=numpy.float32)
    background_factor = 0.1

    # Create input file
    input_filename = tmp_path / "input.h5"
    input_nxdata = {
        **NXDATA_METADATA,
        "q": q,
        "q@units": "A^-1",
        "intensity": input_data,
        "points": [0, 1],
    }
    dicttonx(input_nxdata, input_filename, "/entry/process/integrated")

    # Create background file
    bg_filename = tmp_path / "background.h5"
    bg_nxdata = {
        **NXDATA_METADATA,
        "q": q,
        "q@units": "A^-1",
        "intensity": background_data,
    }
    dicttonx(bg_nxdata, bg_filename, "/entry/process/integrated")

    output_nxprocess_url = DataUrl(
        file_path=tmp_path / "output.h5", data_path="/entry/process_subtracted_detector"
    )
    external_filename = tmp_path / "external.h5"
    inputs = {
        "nxdata_url": f"{input_filename}::/entry/process/integrated",
        "output_nxprocess_url": output_nxprocess_url.path(),
        "external_nxprocess_url": f"{external_filename}::/entry/process_subtracted",
        "background_nxdata_url": f"{bg_filename}::/entry/process/integrated",
        "background_factor": background_factor,
    }
    task = SubtractBackgroundPattern(inputs={**inputs})
    task.execute()

    assert (
        task.get_output_value("nxdata_url")
        == f"{external_filename}::/entry/process_subtracted/integrated"
    )

    with h5py.File(input_filename, "r") as input_file:
        input_nxdata = input_file["/entry/process/integrated"]
        with h5py.File(output_nxprocess_url.file_path(), "r") as output_file:
            output_nxprocess = output_file[output_nxprocess_url.data_path()]
            output_nxdata = output_nxprocess["integrated"]
            numpy.testing.assert_allclose(
                output_nxdata["intensity"][()],
                numpy.ones((2, 100), dtype=numpy.float32) * (2 - background_factor),
            )
            numpy.testing.assert_allclose(input_nxdata["q"][()], output_nxdata["q"][()])
            numpy.testing.assert_allclose(
                input_nxdata["points"][()], output_nxdata["points"][()]
            )
            for name, value in inputs.items():
                stored_value = output_nxprocess["configuration"][name][()]
                if isinstance(stored_value, bytes):
                    stored_value = stored_value.decode()
                assert stored_value == value


@pytest.mark.skipif(not is_silx_compatible(), reason="Silx is too old")
def test_execution_with_errors(tmp_path):
    q = numpy.linspace(0.01, 10, 100, dtype=numpy.float64)
    input_data = 2 * numpy.ones((2, 100), dtype=numpy.float32)
    input_errors = 0.2 * numpy.ones((2, 100), dtype=numpy.float32)
    background_data = numpy.ones((1, 100), dtype=numpy.float32)
    background_errors = 0.1 * numpy.ones((1, 100), dtype=numpy.float32)
    background_factor = 0.5

    # Create input file
    input_filename = tmp_path / "input.h5"
    input_nxdata = {
        **NXDATA_METADATA,
        "q": q,
        "q@units": "A^-1",
        "intensity": input_data,
        "intensity_errors": input_errors,
        "points": [0, 1],
    }
    dicttonx(input_nxdata, input_filename, "/entry/process/integrated")

    # Create background file
    bg_filename = tmp_path / "background.h5"
    bg_nxdata = {
        **NXDATA_METADATA,
        "q": q,
        "q@units": "A^-1",
        "intensity": background_data,
        "intensity_errors": background_errors,
    }
    dicttonx(bg_nxdata, bg_filename, "/entry/process/integrated")

    inputs = {
        "nxdata_url": f"{input_filename}::/entry/process/integrated",
        "output_nxprocess_url": f"{input_filename}::/entry/process_subtracted",
        "background_nxdata_url": f"{bg_filename}::/entry/process/integrated",
        "background_factor": background_factor,
    }
    task = SubtractBackgroundPattern(inputs={**inputs})
    task.execute()

    assert (
        task.get_output_value("nxdata_url")
        == f"{input_filename}::/entry/process_subtracted/integrated"
    )

    with h5py.File(input_filename, "r") as h5file:
        subtracted_intensity = h5file["/entry/process_subtracted/integrated/intensity"][
            ()
        ]
        numpy.testing.assert_allclose(
            subtracted_intensity,
            numpy.ones((2, 100), dtype=numpy.float32) * (2 - background_factor),
        )
        errors = h5file["/entry/process_subtracted/integrated/intensity_errors"][()]
        numpy.testing.assert_allclose(
            errors,
            numpy.sqrt(input_errors**2 + background_factor**2 * background_errors**2),
        )
        numpy.testing.assert_allclose(
            h5file["/entry/process/integrated/q"][()],
            h5file["/entry/process_subtracted/integrated/q"][()],
        )
        numpy.testing.assert_allclose(
            h5file["/entry/process/integrated/points"][()],
            h5file["/entry/process_subtracted/integrated/points"][()],
        )
        for name, value in inputs.items():
            stored_value = h5file["/entry/process_subtracted/configuration"][name][()]
            if isinstance(stored_value, bytes):
                stored_value = stored_value.decode()
            assert stored_value == value


@pytest.mark.skipif(not is_silx_compatible(), reason="Silx is too old")
def test_execution_with_errors_only_on_input(tmp_path):
    q = numpy.linspace(0.01, 10, 100, dtype=numpy.float64)
    input_data = 2 * numpy.ones((2, 100), dtype=numpy.float32)
    input_errors = 0.2 * numpy.ones((2, 100), dtype=numpy.float32)
    background_data = numpy.ones((1, 100), dtype=numpy.float32)
    background_factor = 0.5

    # Create input file
    input_filename = tmp_path / "input.h5"
    input_nxdata = {
        **NXDATA_METADATA,
        "q": q,
        "q@units": "A^-1",
        "intensity": input_data,
        "intensity_errors": input_errors,
        "points": [0, 1],
    }
    dicttonx(input_nxdata, input_filename, "/entry/process/integrated")

    # Create background file
    bg_filename = tmp_path / "background.h5"
    bg_nxdata = {
        **NXDATA_METADATA,
        "q": q,
        "q@units": "A^-1",
        "intensity": background_data,
    }
    dicttonx(bg_nxdata, bg_filename, "/entry/process/integrated")

    inputs = {
        "nxdata_url": f"{input_filename}::/entry/process/integrated",
        "output_nxprocess_url": f"{input_filename}::/entry/process_subtracted",
        "background_nxdata_url": f"{bg_filename}::/entry/process/integrated",
        "background_factor": background_factor,
    }
    task = SubtractBackgroundPattern(inputs={**inputs})
    task.execute()

    with h5py.File(input_filename, "r") as h5file:
        assert "/entry/process_subtracted/integrated/intensity_errors" not in h5file


@pytest.mark.skipif(not is_silx_compatible(), reason="Silx is too old")
def test_with_different_shape_background(tmp_path):
    input_q = numpy.linspace(0.01, 10, 100, dtype=numpy.float64)
    input_data = 2 * numpy.ones((2, 100), dtype=numpy.float32)

    input_filename = tmp_path / "input.h5"
    input_nxdata = {
        **NXDATA_METADATA,
        "q": input_q,
        "q@units": "A^-1",
        "intensity": input_data,
    }
    dicttonx(input_nxdata, input_filename, "/entry/integrated")

    background_q = numpy.linspace(0.01, 10, 50, dtype=numpy.float64)
    background_data = 2 * numpy.ones((2, 50), dtype=numpy.float32)
    bg_filename = tmp_path / "background.h5"
    bg_nxdata = {
        **NXDATA_METADATA,
        "q": background_q,
        "q@units": "A^-1",
        "intensity": background_data,
    }
    dicttonx(bg_nxdata, bg_filename, "/entry/integrated")

    task = SubtractBackgroundPattern(
        inputs={
            "nxdata_url": f"{input_filename}::/entry/integrated",
            "output_nxprocess_url": f"{input_filename}::/entry/process_subtracted",
            "background_nxdata_url": f"{bg_filename}::/entry/integrated",
        }
    )
    with pytest.raises(RuntimeError) as exception:
        task.execute()
    original_exception = exception.value.__cause__
    assert isinstance(original_exception, ValueError)
    assert "not compatible with data pattern" in str(original_exception)


@pytest.mark.skipif(not is_silx_compatible(), reason="Silx is too old")
def test_with_different_unit_background(tmp_path):
    q = numpy.linspace(0.01, 10, 100, dtype=numpy.float64)
    input_data = 2 * numpy.ones((2, 100), dtype=numpy.float32)
    background_data = numpy.ones((1, 100), dtype=numpy.float32)

    # Create input file
    input_filename = tmp_path / "input.h5"
    input_nxdata = {
        **NXDATA_METADATA,
        "q": q,
        "q@units": "A^-1",
        "intensity": input_data,
    }
    dicttonx(input_nxdata, input_filename, "/entry/integrated")

    # Create background file
    bg_filename = tmp_path / "background.h5"
    bg_nxdata = {
        **NXDATA_METADATA,
        "q": q,
        "q@units": "nm^-1",
        "intensity": background_data,
    }
    dicttonx(bg_nxdata, bg_filename, "/entry/integrated")

    task = SubtractBackgroundPattern(
        inputs={
            "nxdata_url": f"{input_filename}::/entry/integrated",
            "output_nxprocess_url": f"{input_filename}::/entry/process_subtracted",
            "background_nxdata_url": f"{bg_filename}::/entry/integrated",
        }
    )
    with pytest.raises(RuntimeError) as exception:
        task.execute()
    original_exception = exception.value.__cause__
    assert isinstance(original_exception, ValueError)
    assert "not compatible with data pattern" in str(original_exception)



import json
import logging
from pathlib import Path
from typing import Union

from ewokscore import Task
from pyFAI import method_registry
from pyFAI.io import ponifile

from .utils import pyfai_utils
from .utils import xrpd_utils

__all__ = ["PyFaiConfig"]

logger = logging.getLogger(__name__)


class PyFaiConfig(
    Task,
    optional_input_names=[
        "energy",
        "geometry",
        "detector",
        "detector_config",
        "mask",
        "flatfield",
        "darkcurrent",
        "integration_options",
        "filenames",
        "filename",
        "calibrant",
        "darkflatmethod",
        "show_merge_warnings",
    ],
    output_names=[
        "energy",
        "geometry",
        "detector",
        "detector_config",
        "mask",
        "flatfield",
        "darkcurrent",
        "integration_options",
        "calibrant",
    ],
):
    """Parse pyFAI calibration and integration parameters.

    Optional inputs:
    - energy (float|None): Energy in KeV (priority 1)
    - geometry (dict|None): pyFAI geometry information (priority 1)
    - detector (str|None): Name of the detector (priority 1)
    - detector_config (dict|None): Configuration of the detector (priority 1)
    - mask (str|numpy.ndarray|None): Filename or data of the detector mask (priority 1)
    - flatfield (str|numpy.ndarray|None): Filename or data of the detector flat-field (priority 1)
    - darkcurrent (str|numpy.ndarray|None): Filename or data of the detector dark-current (priority 1)
    - integration_options (dict|None): Extra pyFAI worker or integration parameters (priority 2)
    - filenames (Sequence[str]|None): PyFAI poni or json file name (priority 3, last file has highest priority)
    - filename (str|None): PyFAI poni or json file name (priority 4)
    - calibrant (str|None): Calibrant name
    - darkflatmethod (str|None): Dark and flat-field correction method

    Outputs:
    - energy (float): Energy in KeV
    - geometry (dict): pyFAI geometry information
    - detector (str): Name of the detector
    - detector_config (dict): Configuration of the detector
    - mask (str|numpy.ndarray|None): Filename or data of the detector mask
    - flatfield (str|numpy.ndarray|None): Filename or data of the detector flat-field
    - darkcurrent (str|numpy.ndarray|None): Filename or data of the detector dark-current
    - integration_options (dict): Extra pyFAI worker or integration parameters
    - calibrant (str|None): Calibrant name
    """

    def run(self):
        input_values = self.get_input_values()
        merged_options = self.merged_integration_options()

        ##########################################################################
        # Extract poni variables to energy, detector, detector_config and geometry
        ##########################################################################

        if "poni" in merged_options and merged_options.get("version", 0) > 3:
            merged_options.update(merged_options.pop("poni"))

        # energy > merged_options["energy"] > merged_options["wavelength"]
        energy = input_values.get("energy", merged_options.pop("energy", None))
        wavelength = merged_options.pop("wavelength", None)
        if energy is None and wavelength is not None:
            energy = xrpd_utils.energy_wavelength(wavelength)

        # detector > merged_options["detector"]
        detector = merged_options.pop("detector", None)
        if not self.missing_inputs.detector:
            detector = input_values["detector"]

        # detector_config > merged_options["detector_config"]
        detector_config = merged_options.pop("detector_config", None)
        if not self.missing_inputs.detector_config:
            detector_config = input_values["detector_config"]

        geometry = {
            k: merged_options.pop(k)
            for k in ["dist", "poni1", "poni2", "rot1", "rot2", "rot3"]
            if k in merged_options
        }
        if not self.missing_inputs.geometry:
            geometry = input_values["geometry"]

        ############################################
        # Extract image correction related variables
        ############################################

        mask = input_values.get("mask", None)
        flatfield = input_values.get("flatfield", None)
        darkcurrent = input_values.get("darkcurrent", None)
        if not self.missing_inputs.darkflatmethod:
            merged_options["darkflatmethod"] = self.inputs.darkflatmethod

        #################################
        # Normalize error model variables
        #################################

        do_poisson = merged_options.pop("do_poisson", None)
        do_azimuthal_error = merged_options.pop("do_azimuthal_error", None)
        error_model = merged_options.pop("error_model", None)
        if not error_model:
            if do_poisson:
                error_model = "poisson"
            if do_azimuthal_error:
                error_model = "azimuthal"
        if error_model:
            merged_options["error_model"] = error_model

        #######################################
        # Check method and integrator function
        #######################################

        method = merged_options.get("method") or ""
        if not isinstance(method, str):
            method = "_".join([_ for _ in method if isinstance(_, str)])
        pmethod = method_registry.IntegrationMethod.parse(method)

        integrator_name = merged_options.get("integrator_name", "")
        if integrator_name in ("sigma_clip", "_sigma_clip_legacy"):
            logger.warning(
                "'%s' is not compatible with the pyfai worker: use 'sigma_clip_ng'",
                integrator_name,
            )
            merged_options["integrator_name"] = "sigma_clip_ng"
        if "sigma_clip_ng" == integrator_name:
            if pmethod and pmethod.split != "no":
                raise ValueError(
                    "to combine sigma clipping with pixel splitting, use 'sigma_clip_legacy'"
                )

        ################################
        # Extract calibration parameters
        ################################

        calibrant = input_values.get("calibrant", None)

        ################################
        # Extract schema versions
        ################################

        _ = merged_options.pop("poni_version", None)

        # There is also "version" which is the JSON schema version.

        ##########
        # Output
        ##########

        self.outputs.energy = energy
        self.outputs.geometry = geometry
        self.outputs.detector = detector
        self.outputs.detector_config = detector_config
        self.outputs.calibrant = calibrant
        self.outputs.mask = mask
        self.outputs.flatfield = flatfield
        self.outputs.darkcurrent = darkcurrent
        self.outputs.integration_options = merged_options

    def _update_integration_options(
        self, current_options: dict, new_options_source: Union[str, dict]
    ):
        if isinstance(new_options_source, dict):
            new_options = new_options_source
        else:
            new_options = pyfai_utils.read_config(new_options_source)

        if self.get_input_value("show_merge_warnings", False):
            common_keys = current_options.keys() & new_options.keys()
            for key in common_keys:
                logger.warning(
                    f"New value of '{key}' ({new_options[key]}) from {new_options_source} will overwrite the current value ({current_options[key]})!"
                )

        current_options.update(new_options)

    def merged_integration_options(self) -> dict:
        """Merge integration options in this order of priority:

        - filename (lowest priority)
        - filenames[0]
        - filenames[1]
        - ...
        - integration_options (highest priority)
        """
        merged_options = dict()
        filenames = list()
        if self.inputs.filename:
            filenames.append(self.inputs.filename)
        if self.inputs.filenames:
            filenames.extend(self.inputs.filenames)
        for filename in filenames:
            self._update_integration_options(merged_options, filename)
        if self.inputs.integration_options:
            self._update_integration_options(
                merged_options,
                pyfai_utils.normalize_parameters(self.inputs.integration_options),
            )
        return merged_options


class SavePyFaiConfig(
    Task,
    input_names=[
        "output_filename",
        "energy",
        "geometry",
        "detector",
    ],
    optional_input_names=[
        "mask",
        "detector_config",
        "integration_options",
    ],
    output_names=["filename"],
):
    """Save inputs as pyFAI calibration and integration configuration file (.json)

    The configuration is saved as a JSON file following pyFAI configuration format.

    Required inputs:
    - output_filename (str): Name of the file where to save pyFAI configuration. Must include the extension
    - energy (float): Energy in KeV
    - geometry (dict): pyFAI geometry information (poni)
    - detector (str): Name of the detector

    Optional inputs:
    - mask (str): Filename of the mask to used
    - detector_config (dict): Configuration of the detector
    - integration_options (dict): Extra configuration fields

    Outputs:
    - filename (str): Saved filename, same as output_filename
    """

    def run(self):
        integration_options = pyfai_utils.normalize_parameters(
            self.get_input_value("integration_options", {})
        )
        version = integration_options.pop("version", 3)

        config = {
            "application": "pyfai-integrate",
            "version": version,
        }

        poni = _create_ponifile(
            self.inputs.energy,
            self.inputs.geometry,
            self.inputs.detector,
            self.get_input_value("detector_config", {}),
        ).as_dict()
        _ = poni.pop("poni_version", None)

        if version >= 4:
            config["poni"] = poni
        else:
            config.update(poni)

        mask = self.get_input_value("mask", None)
        if mask is not None:
            config["do_mask"] = True
            config["mask_file"] = mask

        for key, value in integration_options.items():
            config.setdefault(key, value)  # Do not override already set keys

        filepath = Path(self.inputs.output_filename).absolute()
        filepath.parent.mkdir(parents=True, exist_ok=True)
        filepath.write_text(json.dumps(config, indent=4))

        self.outputs.filename = str(filepath)


class SavePyFaiPoniFile(
    Task,
    input_names=[
        "output_filename",
        "energy",
        "geometry",
        "detector",
    ],
    optional_input_names=[
        "detector_config",
    ],
    output_names=["filename"],
):
    """Save inputs as pyFAI PONI file

    Required inputs:
    - output_filename (str): Name of the file where to save pyFAI PONI. Must include extension.
    - energy (float): Energy in KeV
    - geometry (dict): pyFAI geometry information (poni)
    - detector (str): Name of the detector

    Optional inputs:
    - detector_config (dict): Configuration of the detector

    Outputs:
    - filename (str): Saved filename, same as output_filename
    """

    def run(self):
        poni = _create_ponifile(
            self.inputs.energy,
            self.inputs.geometry,
            self.inputs.detector,
            self.get_input_value("detector_config", {}),
        )

        filepath = Path(self.inputs.output_filename).absolute()
        filepath.parent.mkdir(parents=True, exist_ok=True)
        with filepath.open("w", encoding="ascii") as fd:
            poni.write(fd)

        self.outputs.filename = str(filepath)


def _create_ponifile(
    energy: float,
    geometry: dict,
    detector: str,
    detector_config: dict,
) -> ponifile.PoniFile:
    return ponifile.PoniFile(
        {
            **geometry,  # First so other fields overrides it
            "detector": detector,
            "detector_config": detector_config,
            "wavelength": xrpd_utils.energy_wavelength(energy),
        }
    )



from typing import Dict
from typing import List
from typing import Mapping
from typing import Optional
from typing import Sequence
from typing import Tuple

from AnyQt import QtWidgets
from ewoksdata.data import bliss
from silx.gui.plot import PlotWidget
from silx.gui.plot import ScatterView

from ewoksxrpd.gui import plots
from ewoksxrpd.gui.forms import input_parameters_calibratesingle
from ewoksxrpd.gui.forms import output_parameters_calibratesingle
from ewoksxrpd.gui.forms import pack_geometry
from ewoksxrpd.gui.forms import unpack_enabled_geometry
from ewoksxrpd.gui.forms import unpack_geometry
from ewoksxrpd.gui.trigger_widget import OWTriggerWidget
from ewoksxrpd.tasks.calibrate import CalibrateSingle
from ewoksxrpd.tasks.utils import data_utils

__all__ = ["OWCalibrateSingle"]


class OWCalibrateSingle(OWTriggerWidget, ewokstaskclass=CalibrateSingle):
    name = "CalibrateSingle"
    description = "Single distance calibration"
    icon = "icons/widget.png"
    want_main_area = True

    def __init__(self, *args, **kwargs) -> None:
        self._tabs = QtWidgets.QTabWidget()
        self._show = {
            "input_rings": True,
            "output_rings": False,
            "detected_rings": False,
        }
        self._legends = dict()
        self._cache = dict()
        super().__init__(*args, **kwargs)

    def _init_forms(self):
        parameter_info = input_parameters_calibratesingle(
            self.get_default_input_values()
        )
        self._create_input_form(parameter_info)
        parameter_info = output_parameters_calibratesingle()
        self._create_output_form(parameter_info)

    def _init_control_area(self):
        super()._init_control_area()
        layout = self._get_control_layout()

        w = QtWidgets.QPushButton("Accept refined")
        layout.addWidget(w)
        w.released.connect(self._accept_refined_parameters)

        self._show_input_rings_widget = w = QtWidgets.QCheckBox("Guess rings")
        w.setChecked(self._show["input_rings"])
        layout.addWidget(w)
        w.released.connect(self._set_show_input_rings)

        self._show_output_rings_widget = w = QtWidgets.QCheckBox("Refined rings")
        w.setChecked(self._show["output_rings"])
        layout.addWidget(w)
        w.released.connect(self._set_show_output_rings)

        self._show_detected_rings_widget = w = QtWidgets.QCheckBox("Detected rings")
        w.setChecked(self._show["detected_rings"])
        layout.addWidget(w)
        w.released.connect(self._set_show_detected_rings)

    def _set_show_input_rings(self):
        self._show["input_rings"] = self._show_input_rings_widget.isChecked()
        self._refresh_non_form_input_widgets()

    def _set_show_output_rings(self):
        self._show["output_rings"] = self._show_output_rings_widget.isChecked()
        self._refresh_non_form_output_widgets()

    def _set_show_detected_rings(self):
        self._show["detected_rings"] = self._show_detected_rings_widget.isChecked()
        self._refresh_non_form_output_widgets()

    def _init_main_area(self):
        layout = self._get_main_layout()
        plot = ScatterView()
        w = plot.getPlotWidget()
        w.setGraphXLabel("Dim 2 (pixels)")
        w.setGraphYLabel("Dim 1 (pixels)")
        self._tabs.addTab(plot, "Image")
        layout.addWidget(self._tabs)
        super()._init_main_area()

    def _add_output_form_widget(self):
        self._tabs.addTab(self._output_form, "Refined Geometry")

    def _refresh_non_form_input_widgets(self):
        with self._capture_errors():
            super()._refresh_non_form_input_widgets()
            self._refresh_input_plots()

    def _refresh_non_form_output_widgets(self):
        with self._capture_errors():
            super()._refresh_non_form_output_widgets()
            self._refresh_output_plots()

    def _accept_refined_parameters(self):
        energy = self.get_task_output_value("energy")
        if energy:
            self.update_default_inputs(energy=energy)
        geometry = self.get_task_output_value("geometry")
        if geometry:
            self.update_default_inputs(geometry=dict(geometry))
        self._refresh_input_widgets()

    def _input_form_edited(self):
        super()._input_form_edited()
        fixed = [
            k for k, v in self._input_form.get_parameters_checked().items() if not v
        ]
        self.update_default_inputs(fixed=fixed)
        self._refresh_non_form_input_widgets()

    def _refresh_input_plots(self):
        if self.plot is None:
            return None
        inputs = self.get_task_input_values()
        self._update_image(inputs)
        self._update_input_rings(inputs)

    def _refresh_output_plots(self):
        if self.plot is None:
            return None
        outputs = self.get_task_output_values()
        self._update_detected_rings(outputs)
        inoutputs = {**self.get_task_input_values(), **outputs}
        self._update_output_rings(inoutputs)

    def _update_image(self, inputs):
        image = inputs.get("image")
        if isinstance(image, str):
            previous_image_url = self._cache.get("image")
            if previous_image_url == image:
                return
            self._cache["image"] = image
        self._remove_from_plot("image")
        if not data_utils.is_data(image):
            return
        image = bliss.get_image(image, retry_timeout=0)
        self._legends["image"] = [plots.plot_image(self.plot, image, legend="image")]

    def _update_input_rings(self, values):
        self._remove_from_plot("input_rings")
        if not self._show["input_rings"]:
            return
        self._legends["input_rings"] = self._update_theoretical_rings(values, "input")

    def _update_output_rings(self, values):
        self._remove_from_plot("output_rings")
        if not self._show["output_rings"]:
            return
        self._legends["output_rings"] = self._update_theoretical_rings(values, "output")

    def _update_theoretical_rings(self, values, legend) -> List[str]:
        energy = values.get("energy")
        geometry = values.get("geometry")
        detector = values.get("detector")
        detector_config = values.get("detector_config")
        calibrant = values.get("calibrant")
        max_rings = values.get("max_rings")
        if isinstance(max_rings, Sequence):
            max_rings = max_rings[-1]
        if not energy or not geometry or not detector or not calibrant:
            return list()
        geometry = data_utils.data_from_storage(geometry)
        return plots.plot_theoretical_rings(
            self.plot,
            detector,
            calibrant,
            energy,
            geometry,
            detector_config=detector_config,
            max_rings=max_rings,
            legend=legend,
        )

    def _update_detected_rings(self, values):
        self._remove_from_plot("detected_rings")
        if not self._show["detected_rings"]:
            return
        rings = values.get("rings")
        if not rings:
            return
        rings = data_utils.data_from_storage(rings)
        self._legends["detected_rings"] = plots.plot_detected_rings(self.plot, rings)

    def _remove_from_plot(self, name: str) -> None:
        legends = self._legends.pop(name, list())
        for legend in legends:
            self.plot.remove(legend=legend)

    def _values_from_form(
        self, values: Mapping, checked: Dict[str, bool], output: bool = False
    ) -> Mapping:
        return pack_geometry(values, checked)

    def _values_to_form(
        self, values: Mapping, output: bool = False
    ) -> Tuple[Mapping, Dict[str, bool]]:
        return unpack_geometry(values)

    def _enabled_to_form(
        self, enabled: Dict[str, bool], output: bool = False
    ) -> Dict[str, bool]:
        return unpack_enabled_geometry(enabled)

    @property
    def plot(self) -> Optional[PlotWidget]:
        if self._tabs.count() == 0:
            return None
        return self._tabs.widget(0).getPlotWidget()



from contextlib import contextmanager
from importlib.metadata import version
from typing import Iterator
from typing import Union

import h5py
import numpy
from ewokscore import TaskWithProgress
from ewoksdata.data import bliss
from ewoksdata.data import nexus
from ewoksdata.data.contextiterator import contextiterator
from packaging.version import Version
from silx.io import h5py_utils
from silx.io.url import DataUrl

from .utils import data_utils

_LIMA_TEMPLATE_SUPPORTED = Version(version("blissdata")) >= Version("1.1.0")
_PRIORITIZE_NONNATIVEH5ITEMS_SUPPORTED = Version(version("blissdata")) >= Version(
    "2.0.0"
)


class TaskWithDataAccess(
    TaskWithProgress,
    optional_input_names=[
        "retry_timeout",
        "retry_period",
        "lima_url_template",
        "lima_url_template_args",
        "prioritize_non_native_h5items",
    ],
    register=False,
):
    def get_retry_options(self):
        retry_timeout = self.get_input_value("retry_timeout", None)
        retry_period = self.get_input_value("retry_period", None)
        return {"retry_timeout": retry_timeout, "retry_period": retry_period}

    def _get_blissdata_options(self):
        lima_url_template = self.get_input_value("lima_url_template", None)
        lima_url_template_args = self.get_input_value("lima_url_template_args", None)
        prioritize_non_native_h5items = self.get_input_value(
            "prioritize_non_native_h5items", False
        )
        if _LIMA_TEMPLATE_SUPPORTED:
            blissdata_options = {
                "lima_url_template": lima_url_template,
                "lima_url_template_args": lima_url_template_args,
                **self.get_retry_options(),
            }
            if _PRIORITIZE_NONNATIVEH5ITEMS_SUPPORTED:
                blissdata_options["prioritize_non_native_h5items"] = (
                    prioritize_non_native_h5items
                )
            return blissdata_options
        if lima_url_template:
            raise ValueError("'lima_url_template' requires blissdata>=1.1.0")
        if lima_url_template_args:
            raise ValueError("'lima_url_template_args' requires blissdata>=1.1.0")
        if prioritize_non_native_h5items:
            raise ValueError(
                "'prioritize_non_native_h5items' requires blissdata>=2.0.0"
            )
        return self.get_retry_options()

    @contextmanager
    def open_h5item(
        self,
        url: Union[str, DataUrl],
        create: bool = False,
        overwrite: bool = False,
        **openoptions,
    ) -> Iterator[Union[h5py.Group, h5py.Dataset, numpy.ndarray]]:
        if not isinstance(url, DataUrl):
            url = DataUrl(url)
        retryoptions = self.get_retry_options()
        if create:
            url = nexus.create_url(url, overwrite=overwrite, **retryoptions)
        with h5py_utils.open_item(
            url.file_path(), url.data_path(), **retryoptions, **openoptions
        ) as item:
            idx = url.data_slice()
            if idx is None:
                yield item
            else:
                yield item[idx]

    def get_data(self, *args, **kw):
        kw.update(self.get_retry_options())
        return bliss.get_data(*args, **kw)

    def get_image(self, *args, **kw):
        kw.update(self.get_retry_options())
        return bliss.get_image(*args, **kw)

    @contextiterator
    def iter_bliss_data(self, *args, **kw):
        kw.update(self._get_blissdata_options())
        with bliss.iter_bliss_scan_data(*args, **kw) as iterator:
            yield from iterator

    def iter_bliss_data_from_memory(self, *args, **kw):
        kw.update(self.get_retry_options())
        yield from bliss.iter_bliss_scan_data_from_memory(*args, **kw)

    def iter_bliss_scan_data_from_memory_slice(self, *args, **kw):
        kw.update(self.get_retry_options())
        yield from bliss.iter_bliss_scan_data_from_memory_slice(*args, **kw)

    def link_bliss_scan(self, *args, **kw):
        kw.update(self._get_blissdata_options())
        return data_utils.link_bliss_scan(*args, **kw)



from ewoksorange.gui.owwidgets.nothread import OWEwoksWidgetNoThread
from ewoksorange.gui.widgets.parameter_form import ParameterForm

from ewoksxrpd.tasks.pyfaiconfig import SavePyFaiPoniFile

__all__ = ["OWSavePyFaiPoniFile"]


class OWSavePyFaiPoniFile(OWEwoksWidgetNoThread, ewokstaskclass=SavePyFaiPoniFile):
    name = "SavePyFaiPoniFile"
    description = "Save pyFAI PONI file"
    icon = "icons/widget.png"
    want_main_area = False

    def __init__(self):
        super().__init__()

        self._parameter_form = ParameterForm(parent=self.controlArea)
        self._parameter_form.addParameter(
            "output_filename",
            label="pyFAI PONI filename (*.poni)",
            value_for_type="",
            select="file",
            value_change_callback=self._inputs_changed,
        )
        self._update_parameter_values()

        self.controlArea.layout().addStretch(1)

    def _inputs_changed(self):
        new_values = self._parameter_form.get_parameter_values()
        self.update_default_inputs(**new_values)

    def _update_parameter_values(self):
        initial_values = self.get_default_input_values()
        self._parameter_form.set_parameter_values(initial_values)
